@@ -9,12 +9,6 @@
 // If this gets raised to a double digit number, the changes introduced by commit f24988ab68a94 will need to be modified to accommodate for that
 #define DEV_MAX     9
 extern usbdevice keyboard[DEV_MAX];
-// Is a device active?
-#ifdef OS_LINUX
-#define IS_CONNECTED(kb) ((kb) && (kb)->handle && (((kb)->uinput_kb && (kb)->uinput_mouse) || IS_SINGLE_EP(kb)))
-#else
-#define IS_CONNECTED(kb) ((kb) && (kb)->handle && (kb)->event)
-#endif
 
 #ifdef NO_FAIR_MUTEX_QUEUEING
 #define QUEUED_MUTEX_INITIALIZER PTHREAD_MUTEX_INITIALIZER
@@ -66,6 +60,10 @@ extern pthread_mutex_t interruptmutex[DEV_MAX];
 extern pthread_cond_t interruptcond[DEV_MAX];
 #define intcond(kb) (interruptcond + INDEX_OF(kb, keyboard))
 
+// Mutex to access the children for each device
+extern pthread_mutex_t childrenmutex[DEV_MAX];
+#define cmutex(kb) (childrenmutex + INDEX_OF(kb, keyboard))
+
 int init_cond_monotonic(void);
 
 // Sets up device hardware, after software initialization is finished. Also used during resets
@@ -73,6 +71,9 @@ int init_cond_monotonic(void);
 int start_dev(usbdevice* kb, int makeactive);
 int start_kb_legacy(usbdevice* kb, int makeactive);
 int start_mouse_legacy(usbdevice* kb, int makeactive);
+int start_mouse_bragi(usbdevice* kb, int makeactive);
+int start_keyboard_bragi(usbdevice* kb, int makeactive);
+int start_dongle_bragi(usbdevice* kb, int makeactive);
 
 // Activates/deactives software control on a device. Return 0 on success
 int setactive_kb(usbdevice* kb, int active);
@@ -81,17 +82,20 @@ int setactive_mouse(usbdevice* kb, int active);
 /// \brief setactive() calls via the corresponding kb->vtable either the active() or the idle() function.
 /// \n active() is called if the parameter makeactive is true, idle if it is false.
 /// \n What function is called effectively is device dependent. Have a look at \a device_vtable.c for more information.
-#define setactive(kb, makeactive) ((makeactive) ? (kb)->vtable->active((kb), 0, 0, 0, 0) : (kb)->vtable->idle((kb), 0, 0, 0, 0))
+#define setactive(kb, makeactive) ((makeactive) ? (kb)->vtable.active((kb), 0, 0, 0, 0) : (kb)->vtable.idle((kb), 0, 0, 0, 0))
 
 // Command: Activate a device
 int cmd_active_kb(usbdevice* kb, usbmode* dummy1, int dummy2, int dummy3, const char* dummy4);
 int cmd_active_mouse(usbdevice* kb, usbmode* dummy1, int dummy2, int dummy3, const char* dummy4);
+int cmd_active_bragi(usbdevice* kb, usbmode* dummy1, int dummy2, int dummy3, const char* dummy4);
 // Command: Deactivate a device
 int cmd_idle_kb(usbdevice* kb, usbmode* dummy1, int dummy2, int dummy3, const char* dummy4);
 int cmd_idle_mouse(usbdevice* kb, usbmode* dummy1, int dummy2, int dummy3, const char* dummy4);
+int cmd_idle_bragi(usbdevice* kb, usbmode* dummy1, int dummy2, int dummy3, const char* dummy4);
 // Command: Set poll rate
 int cmd_pollrate(usbdevice* kb, usbmode* dummy1, int dummy2, int rate, const char* dummy3);
 int cmd_pollrate_legacy(usbdevice* kb, usbmode* dummy1, int dummy2, int rate, const char* dummy3);
+int cmd_pollrate_bragi(usbdevice* kb, usbmode* dummy1, int dummy2, int rate, const char* dummy3);
 
 // Sets a device's current mode index. This is only used on the legacy K95; the RGB keyboards have no gettable HW index.
 void setmodeindex_legacy(usbdevice* kb, int index);
@@ -99,6 +103,12 @@ void setmodeindex_legacy(usbdevice* kb, int index);
 // Resets devices using the NXP HID protocol
 void nxp_reset(usbdevice* kb, usbmode* mode, int dummy1, int dummy2, const char* type);
 
+// Used to clear the input struct and force an rgb update when switching between hw and sw modes
+void clear_input_and_rgb(usbdevice* kb, const int active);
+
+void nxp_get_battery_info(usbdevice* kb);
+void bragi_get_battery_info(usbdevice* kb);
+
 // Per-key input settings for device setup
 // The upper nybble controls input mode. 0x80 generates a normal HID interrupt, 0x40 generates a proprietary interrupt. 0xc0 generates both.
 // The exceptions are the proprietary Corsair keys, which only report HID input in BIOS mode and only report Corsair input in non-BIOS mode.
