@@ -5,6 +5,8 @@
 #include "input.h"
 #include "profile.h"
 #include "usb.h"
+#include "nxp_proto.h"
+#include "legacykb_proto.h"
 
 int start_kb_legacy(usbdevice* kb, int makeactive){
     (void)makeactive;
@@ -21,13 +23,7 @@ int setactive_kb(usbdevice* kb, int active){
     if(NEEDS_FW_UPDATE(kb))
         return 0;
 
-    queued_mutex_lock(imutex(kb));
-    kb->active = !!active;
-    kb->profile->lastlight.forceupdate = 1;
-    // Clear input
-    memset(&kb->input.keys, 0, sizeof(kb->input.keys));
-    inputupdate(kb);
-    queued_mutex_unlock(imutex(kb));
+    clear_input_and_rgb(kb, active);
 
     uchar msg[3][MSG_SIZE] = {
         { CMD_SET, FIELD_SPECIAL, 0 },                // Disables or enables HW control for top row
@@ -37,7 +33,7 @@ int setactive_kb(usbdevice* kb, int active){
     if(active){
         // Put the M-keys (K95) as well as the Brightness/Lock keys into software-controlled mode.
         msg[0][2] = MODE_SOFTWARE;
-        if(!usbsend(kb, msg[0], 1))
+        if(!usbsend(kb, msg[0], MSG_SIZE, 1))
             return -1;
         DELAY_MEDIUM(kb);
         // Set input mode on the keys. They must be grouped into packets of 60 bytes (+ 4 bytes header)
@@ -55,20 +51,20 @@ int setactive_kb(usbdevice* kb, int active){
             }
             // Byte 2 = pair count (usually 30, less on final message)
             msg[1][2] = pair;
-            if(!usbsend(kb, msg[1], 1))
+            if(!usbsend(kb, msg[1], MSG_SIZE, 1))
                 return -1;
         }
         // Commit new input settings
-        if(!usbsend(kb, msg[2], 1))
+        if(!usbsend(kb, msg[2], MSG_SIZE, 1))
             return -1;
         DELAY_MEDIUM(kb);
     } else {
         // Set the M-keys back into hardware mode, restore hardware RGB profile. It has to be sent twice for some reason.
         msg[0][2] = MODE_HARDWARE;
-        if(!usbsend(kb, msg[0], 1))
+        if(!usbsend(kb, msg[0], MSG_SIZE, 1))
             return -1;
         DELAY_MEDIUM(kb);
-        if(!usbsend(kb, msg[0], 1))
+        if(!usbsend(kb, msg[0], MSG_SIZE, 1))
             return -1;
         DELAY_MEDIUM(kb);
 #ifdef OS_LINUX
@@ -97,17 +93,17 @@ int setactive_kb(usbdevice* kb, int active){
             }
             // Byte 2 = pair count (usually 30, less on final message)
             msg[1][2] = pair;
-            if(!usbsend(kb, msg[1], 1))
+            if(!usbsend(kb, msg[1], MSG_SIZE, 1))
                 return -1;
         }
         // Commit new input settings
-        if(!usbsend(kb, msg[2], 1))
+        if(!usbsend(kb, msg[2], MSG_SIZE, 1))
             return -1;
         DELAY_MEDIUM(kb);
 #endif
     }
     // Update indicator LEDs if the profile contains settings for them
-    kb->vtable->updateindicators(kb, 0);
+    kb->vtable.updateindicators(kb, 0);
     return 0;
 }
 
