@@ -1,5 +1,7 @@
 #include "dpi.h"
 #include "usb.h"
+#include "nxp_proto.h"
+#include <assert.h>
 
 void cmd_dpi(usbdevice* kb, usbmode* mode, int dummy, const char* stages, const char* values){
     (void)kb;
@@ -67,6 +69,8 @@ void cmd_lift(usbdevice* kb, usbmode* mode, int dummy1, int dummy2, const char*
         return;
     if(heightnum > LIFT_MAX || heightnum < LIFT_MIN)
         return;
+    if(IS_DARK_CORE_NXP(kb) && heightnum < 3)
+        return;
     mode->dpi.lift = heightnum;
 }
 
@@ -108,9 +112,10 @@ int updatedpi(usbdevice* kb, int force){
         return 0;
     dpiset* lastdpi = &kb->profile->lastdpi;
     dpiset* newdpi = &kb->profile->currentmode->dpi;
+    lighting* newlight = &kb->profile->currentmode->light;
     // Don't do anything if the settings haven't changed
-    if(!force && !lastdpi->forceupdate && !newdpi->forceupdate
-            && !memcmp(lastdpi, newdpi, sizeof(dpiset)))
+    if(!(force || lastdpi->forceupdate || newdpi->forceupdate
+            || memcmp(lastdpi, newdpi, sizeof(dpiset))))
         return 0;
     lastdpi->forceupdate = newdpi->forceupdate = 0;
 
@@ -132,7 +137,7 @@ int updatedpi(usbdevice* kb, int force){
                 newenabled = lastdpi->enabled | 1 << newdpi->current;
             }
             uchar data_pkt[MSG_SIZE] = { CMD_SET, FIELD_MOUSE, MOUSE_DPIMASK, 0, newenabled };
-            if(!usbsend(kb, data_pkt, 1))
+            if(!usbsend(kb, data_pkt, sizeof(data_pkt), 1))
                 return -2;
             // Cache the flags we wrote.
             lastdpi->enabled = newenabled;
@@ -149,7 +154,7 @@ int updatedpi(usbdevice* kb, int force){
             data_pkt[6] = (newdpi->x[newdpi->current] >> 8) & 0xFF;
             data_pkt[7] = newdpi->y[newdpi->current] & 0xFF;
             data_pkt[8] = (newdpi->y[newdpi->current] >> 8) & 0xFF;
-            if(!usbsend(kb, data_pkt, 1))
+            if(!usbsend(kb, data_pkt, sizeof(data_pkt), 1))
                 return -1;
             // Set these values in the cache so we don't rewrite them.
             lastdpi->x[newdpi->current] = newdpi->x[newdpi->current];
@@ -157,14 +162,17 @@ int updatedpi(usbdevice* kb, int force){
         }
         // Set current DPI stage.
         uchar data_pkt[MSG_SIZE] = { CMD_SET, FIELD_MOUSE, MOUSE_DPI, 0, newdpi->current };
-        if(!usbsend(kb, data_pkt, 1))
+        if(!usbsend(kb, data_pkt, sizeof(data_pkt), 1))
             return -2;
     }
-    
+
     // Send X/Y DPIs. We've changed to the new stage already so these can be set
     // safely.
+    // The dpi zones for devices in NXP_RGB_IN_DPI_PKT() are last in the keymap
+    // (and in the r, g, b arrays, so we need to be sure we don't go out of bounds.
+    static_assert(DPI_RGB_START + DPI_COUNT - 1 < N_KEYS_EXTENDED, "For loop will attempt to read dpi rgb data out of bounds");
     for(int i = 0; i < DPI_COUNT; i++){
-        if (newdpi->x[i] == lastdpi->x[i] && newdpi->y[i] == lastdpi->y[i] && !force)
+        if (!(newdpi->x[i] != lastdpi->x[i] || newdpi->y[i] != lastdpi->y[i] || force))
             continue;
         uchar data_pkt[MSG_SIZE] = { CMD_SET, FIELD_MOUSE, MOUSE_DPIPROF, 0 };
         data_pkt[2] |= i;
@@ -175,24 +183,32 @@ int updatedpi(usbdevice* kb, int force){
         data_pkt[6] = (newdpi->x[i] >> 8) & 0xFF;
         data_pkt[7] = newdpi->y[i] & 0xFF;
         data_pkt[8] = (newdpi->y[i] >> 8) & 0xFF;
-        if(!usbsend(kb, data_pkt, 1))
+
+
+        if(NXP_RGB_IN_DPI_PKT(kb)) {
+            data_pkt[9] = newlight->r[DPI_RGB_START + i];
+            data_pkt[10] = newlight->g[DPI_RGB_START + i];
+            data_pkt[11] = newlight->b[DPI_RGB_START + i];
+        }
+
+        if(!usbsend(kb, data_pkt, sizeof(data_pkt), 1))
             return -1;
     }
 
     // Send settings
     if (newdpi->enabled != lastdpi->enabled || force) {
         uchar data_pkt[MSG_SIZE] = { CMD_SET, FIELD_MOUSE, MOUSE_DPIMASK, 0, newdpi->enabled };
-        if(!usbsend(kb, data_pkt, 1))
+        if(!usbsend(kb, data_pkt, sizeof(data_pkt), 1))
             return -2;
     }
     if (newdpi->lift != lastdpi->lift || force) {
         uchar data_pkt[MSG_SIZE] = { CMD_SET, FIELD_MOUSE, MOUSE_LIFT, 0, newdpi->lift };
-        if(!usbsend(kb, data_pkt, 1))
+        if(!usbsend(kb, data_pkt, sizeof(data_pkt), 1))
             return -2;
     }
     if (newdpi->snap != lastdpi->snap || force) {
         uchar data_pkt[MSG_SIZE] = { CMD_SET, FIELD_MOUSE, MOUSE_SNAP, 0, newdpi->snap, 0x05 };
-        if(!usbsend(kb, data_pkt, 1))
+        if(!usbsend(kb, data_pkt, sizeof(data_pkt), 1))
             return -2;
     }
 
@@ -214,7 +230,7 @@ int savedpi(usbdevice* kb, dpiset* dpi, lighting* light){
         data_pkt[9] = light->r[LED_MOUSE + N_MOUSE_ZONES + i];
         data_pkt[10] = light->g[LED_MOUSE + N_MOUSE_ZONES + i];
         data_pkt[11] = light->b[LED_MOUSE + N_MOUSE_ZONES + i];
-        if(!usbsend(kb, data_pkt, 1))
+        if(!usbsend(kb, data_pkt, sizeof(data_pkt), 1))
             return -1;
     }
 
@@ -225,7 +241,7 @@ int savedpi(usbdevice* kb, dpiset* dpi, lighting* light){
         { CMD_SET, FIELD_MOUSE, MOUSE_LIFT, 1, dpi->lift },
         { CMD_SET, FIELD_MOUSE, MOUSE_SNAP, 1, dpi->snap, 0x05 }
     };
-    if(!usbsend(kb, data_pkt[0], 4))
+    if(!usbsend(kb, data_pkt[0], MSG_SIZE, 4))
         return -2;
     // Finished
     return 0;
@@ -241,7 +257,7 @@ int loaddpi(usbdevice* kb, dpiset* dpi, lighting* light){
     };
     uchar in_pkt[4][MSG_SIZE];
     for(int i = 0; i < 4; i++){
-        if(!usbrecv(kb, data_pkt[i], in_pkt[i]))
+        if(!usbrecv(kb, data_pkt[i], MSG_SIZE, in_pkt[i]))
             return -2;
         if(memcmp(in_pkt[i], data_pkt[i], 4)){
             ckb_err("Bad input header");
@@ -264,7 +280,7 @@ int loaddpi(usbdevice* kb, dpiset* dpi, lighting* light){
         uchar dpi_pkt[MSG_SIZE] = { CMD_GET, FIELD_MOUSE, MOUSE_DPIPROF, 1 };
         uchar dpi_in_pkt[MSG_SIZE];
         dpi_pkt[2] |= i;
-        if(!usbrecv(kb, dpi_pkt, dpi_in_pkt))
+        if(!usbrecv(kb, dpi_pkt, sizeof(dpi_pkt), dpi_in_pkt))
             return -2;
         if(memcmp(dpi_in_pkt, dpi_pkt, 4)){
             ckb_err("Bad input header");
