@@ -2,6 +2,7 @@
 #include "firmware.h"
 #include "notify.h"
 #include "usb.h"
+#include "nxp_proto.h"
 
 #define FW_OK       0
 #define FW_NOFILE   -1
@@ -21,7 +22,7 @@ int getfwversion(usbdevice* kb){
         kb->layout = LAYOUT_UNKNOWN;
     }
 
-    if(!usbrecv(kb, data_pkt, in_pkt))
+    if(!usbrecv(kb, data_pkt, sizeof(data_pkt), in_pkt))
         return -1;
     if(in_pkt[0] != CMD_GET || in_pkt[1] != FIELD_IDENT){
         ckb_err("Bad input header");
@@ -49,17 +50,18 @@ int getfwversion(usbdevice* kb){
         ckb_warn("Got vendor ID %04x (expected %04x)", vendor, kb->vendor);
     if(product != kb->product)
         ckb_warn("Got product ID %04x (expected %04x)", product, kb->product);
+    if(version != kb->fwversion && kb->fwversion != 0)
+        ckb_warn("Got firmware version %04x (expected %04x)", version, kb->fwversion);
+
     // Set firmware version and poll rate
-    if(version == 0 || bootloader == 0){
-        // Needs firmware update
-        kb->fwversion = 0;
+    kb->fwversion = version;
+    kb->pollrate = poll;
+    kb->bldversion = bootloader;
+    kb->radioappversion = kb->radiobldversion = UINT32_MAX;
+
+    if(version == 0 || bootloader == 0)
         kb->pollrate = -1;
-    } else {
-        if(version != kb->fwversion && kb->fwversion != 0)
-            ckb_warn("Got firmware version %04x (expected %04x)", version, kb->fwversion);
-        kb->fwversion = version;
-        kb->pollrate = poll;
-    }
+
     // Physical layout detection.
     if (kb->layout == LAYOUT_UNKNOWN) {
         kb->layout = in_pkt[23] + 1;
@@ -68,6 +70,41 @@ int getfwversion(usbdevice* kb){
             kb->layout = LAYOUT_UNKNOWN;
         }
     }
+    // Wireless requires extra handshake packets.
+    if(IS_WIRELESS_DEV(kb)){
+        uchar wireless_pkt[5][MSG_SIZE] = {
+            { CMD_GET, 0xae, 0 },
+            { CMD_GET, 0x4a, 0 },
+            { CMD_GET, 0x50, 0 },
+            { CMD_SET, 0xad, 0x00, 0x00, 100 }, // Opacity packet.
+            { CMD_SET, 0xaa, 0 },               // Create blank colour profiles.
+        };
+        if(!usbrecv(kb, wireless_pkt[0], MSG_SIZE, in_pkt))
+            return -1;
+        memcpy(&vendor, in_pkt + 4, 2);
+        memcpy(&product, in_pkt + 6, 2);
+        memcpy(&version, in_pkt + 8, 2);
+        if(vendor != kb->vendor)
+            ckb_warn("Got wireless vendor ID %04x (expected %04x)\n", vendor, kb->vendor);
+        if(product != kb->product)
+            ckb_warn("Got wireless product ID %04x (expected %04x)\n", product, kb->product);
+        // More handshake packets.
+        for(int i = 1; i < 3; i++)
+            if(!usbrecv(kb, wireless_pkt[i], MSG_SIZE, in_pkt))
+                return -1;
+        if(!usbsend(kb, wireless_pkt[3], MSG_SIZE, 1))
+            return -1;
+        // Generate blank colour profiles.
+        for(int profile = 1; profile < 6; profile++) {
+            wireless_pkt[4][5] = 0xff; // Blank profile command.
+            wireless_pkt[4][15] = profile;
+            if(!usbsend(kb, wireless_pkt[4], MSG_SIZE, 1))
+                return -1;
+        }
+        /// FIXME: REMOVE THIS WHEN HARDWARE PROFILES AND WIRELESS FW UPDATE ARE ADDED
+        kb->features &= ~(FEAT_HWLOAD | FEAT_FWUPDATE);
+    }
+
     return 0;
 }
 
@@ -140,14 +177,14 @@ int fwupdate(usbdevice* kb, const char* path, int nnumber){
             }
         }
         if(index == 1){
-            if(!usbsend(kb, data_pkt[0], 1)){
+            if(!usbsend(kb, data_pkt[0], MSG_SIZE, 1)){
                 ckb_err("Firmware update failed");
                 free(fwdata);
                 return FW_USBFAIL;
             }
             // The above packet can take a lot longer to process, so wait for a while
             sleep(3);
-            if(!usbsend(kb, data_pkt[2], npackets - 1)){
+            if(!usbsend(kb, data_pkt[2], MSG_SIZE, npackets - 1)){
                 ckb_err("Firmware update failed");
                 free(fwdata);
                 return FW_USBFAIL;
@@ -156,7 +193,7 @@ int fwupdate(usbdevice* kb, const char* path, int nnumber){
             // If the output ends here, set the length byte appropriately
             if(output >= length)
                 data_pkt[npackets][2] = length - last;
-            if(!usbsend(kb, data_pkt[1], npackets)){
+            if(!usbsend(kb, data_pkt[1], MSG_SIZE, npackets)){
                 ckb_err("Firmware update failed");
                 free(fwdata);
                 return FW_USBFAIL;
@@ -169,7 +206,7 @@ int fwupdate(usbdevice* kb, const char* path, int nnumber){
         { CMD_SET, FIELD_FW_DATA, 0xf0, 0x00, 0x00, 0x00, index },
         { CMD_SET, FIELD_RESET, RESET_SLOW, 0 }
     };
-    if(!usbsend(kb, data_pkt2[0], 2)){
+    if(!usbsend(kb, data_pkt2[0], MSG_SIZE, 2)){
         ckb_err("Firmware update failed");
         free(fwdata);
         return FW_USBFAIL;
