@@ -4,6 +4,7 @@
 #include "notify.h"
 #include "profile.h"
 #include "usb.h"
+#include "nxp_proto.h"
 
 // Define an ordered dithering table by using bit reversion.
 #define BR1(x) ((((x) & 0xaa) >> 1) | (((x) & 0x55) << 1))
@@ -40,7 +41,7 @@ static int update_sidelights(usbdevice* kb) {
     uchar data_pkt[MSG_SIZE] = {
         CMD_SET, FIELD_LIGHTING, MODE_SIDELIGHT, 0, !!newlight->sidelight, 0x00
     };
-    if(!usbsend(kb, data_pkt, 1))
+    if(!usbsend(kb, data_pkt, sizeof(data_pkt), 1))
         return -1;
     return 0;
 }
@@ -94,7 +95,9 @@ static void makergb_k55(const lighting* light, uchar data_pkt[MSG_SIZE]){
 
 static int rgbcmp(const lighting* lhs, const lighting* rhs){
     // Compare two light structures, up until the first three generic zones for the K55.
-    return memcmp(lhs->r, rhs->r, LED_GENERIC_FIRST + 3) || memcmp(lhs->g, rhs->g, LED_GENERIC_FIRST + 3) || memcmp(lhs->b, rhs->b, LED_GENERIC_FIRST + 3);
+    // We compare led scancodes, not keymap array indices
+    // 0xA2 is the highest scancode (topbar16) + 1 because that's an index and we need length
+    return memcmp(lhs->r, rhs->r, 0xA2 + 1) || memcmp(lhs->g, rhs->g, 0xA2 + 1) || memcmp(lhs->b, rhs->b, 0xA2 + 1);
 }
 
 int updatergb_kb(usbdevice* kb, int force){
@@ -108,6 +111,9 @@ int updatergb_kb(usbdevice* kb, int force){
         return 0;
     lastlight->forceupdate = newlight->forceupdate = 0;
 
+    if(IS_K63_WL(kb))
+        return updatergb_wireless(kb, lastlight, newlight);
+
     if (kb->product == P_K66 || kb->product == P_K68_NRGB) {
         // The K68 NRGB doesn't support winlock setting through the
         // normal packets, so we have to use a different packet to set it.
@@ -116,10 +122,10 @@ int updatergb_kb(usbdevice* kb, int force){
         uchar winlock_pkt[MSG_SIZE] = {
             CMD_SET, FIELD_LIGHTING, MODE_WINLOCK, 0, winlock_status, 0
         };
-        if (!usbsend(kb, winlock_pkt, 1))
+        if (!usbsend(kb, winlock_pkt, sizeof(winlock_pkt), 1))
             return -1;
     }
-   
+
     // Don't send data if the keyboard has no lighting.
     if (!HAS_FEATURES(kb, FEAT_RGB))
         return 0;
@@ -128,13 +134,13 @@ int updatergb_kb(usbdevice* kb, int force){
         // The K55 has its own packet type, because it only has three lighting zones.
         uchar data_pkt[MSG_SIZE] = { 0x07, 0x25, 0x00 };
         makergb_k55(newlight, data_pkt);
-        if (!usbsend(kb, data_pkt, 1))
+        if (!usbsend(kb, data_pkt, sizeof(data_pkt), 1))
             return -1;
     } else if(IS_FULLRANGE(kb)) {
         // Update strafe sidelights if necessary
-	if (IS_STRAFE(kb) && update_sidelights(kb))
+        if (IS_STRAFE(kb) && update_sidelights(kb))
             return -1;
-        
+
         // 16.8M color lighting works fine on strafe and is the only way it actually works
         uchar data_pkt[12][MSG_SIZE] = {
             // Red
@@ -160,11 +166,11 @@ int updatergb_kb(usbdevice* kb, int force){
         if(IS_MONOCHROME_DEV(kb))
             data_pkt_count = 4;
 
-        if(!usbsend(kb, data_pkt[0], data_pkt_count))
+        if(!usbsend(kb, data_pkt[0], MSG_SIZE, data_pkt_count))
             return -1;
     } else {
         // Update strafe sidelights if necessary
-	if (IS_STRAFE(kb) && update_sidelights(kb))
+        if (IS_STRAFE(kb) && update_sidelights(kb))
             return -1;
         // On older keyboards it looks flickery and causes lighting glitches, so we don't use it.
         uchar data_pkt[5][MSG_SIZE] = {
@@ -175,7 +181,7 @@ int updatergb_kb(usbdevice* kb, int force){
             { CMD_SET, FIELD_KB_9BCLR, 0x00, 0x00, 0xD8 }
         };
         makergb_512(newlight, data_pkt, kb->dither ? ordered8to3 : quantize8to3);
-        if(!usbsend(kb, data_pkt[0], 5))
+        if(!usbsend(kb, data_pkt[0], MSG_SIZE, 5))
             return -1;
     }
 
@@ -203,11 +209,11 @@ int savergb_kb(usbdevice* kb, lighting* light, int mode){
             { CMD_SET, FIELD_KB_HWCLR, 0x03, 0x01, 0x01, mode + 1, COLOR_BLUE }
         };
         makergb_full(light, data_pkt);
-        if(!usbsend(kb, data_pkt[0], 12))
+        if(!usbsend(kb, data_pkt[0], MSG_SIZE, 12))
             return -1;
         if (IS_STRAFE(kb)){ // end save
             uchar save_end_pkt[MSG_SIZE] = { CMD_SET, FIELD_KB_HWCLR, 0x04, 0x01, 0x01 };
-            if(!usbsend(kb, save_end_pkt, 1))
+            if(!usbsend(kb, save_end_pkt, sizeof(save_end_pkt), 1))
                 return -1;
         }
     } else {
@@ -219,7 +225,7 @@ int savergb_kb(usbdevice* kb, lighting* light, int mode){
             { CMD_SET, FIELD_KB_HWCLR, 0x02, 0x00, 0x01, mode + 1 }
         };
         makergb_512(light, data_pkt, kb->dither ? ordered8to3 : quantize8to3);
-        if(!usbsend(kb, data_pkt[0], 5))
+        if(!usbsend(kb, data_pkt[0], MSG_SIZE, 5))
             return -1;
     }
     return 0;
@@ -264,7 +270,7 @@ int loadrgb_kb(usbdevice* kb, lighting* light, int mode){
         uchar* colors[3] = { light->r, light->g, light->b };
         for(int clr = 0; clr < 3; clr++){
             for(int i = 0; i < 4; i++){
-                if(!usbrecv(kb, data_pkt[i + clr * 4], in_pkt[i]))
+                if(!usbrecv(kb, data_pkt[i + clr * 4], MSG_SIZE, in_pkt[i]))
                     return -1;
 
                 uchar* comparePacket = data_pkt[i + clr * 4];   ///< That is the old comparison method: you get back what you sent.
@@ -286,7 +292,7 @@ int loadrgb_kb(usbdevice* kb, lighting* light, int mode){
                     in_pkt[2][1] = 0x99;
                     in_pkt[2][2] = 0x99;
                     in_pkt[2][3] = 0x99;
-                    usbrecv(kb, in_pkt[2], in_pkt[2]); // just to find it in the wireshark log
+                    usbrecv(kb, in_pkt[2], MSG_SIZE, in_pkt[2]); // just to find it in the wireshark log
                     return -1;
                 }
             }
@@ -310,11 +316,11 @@ int loadrgb_kb(usbdevice* kb, lighting* light, int mode){
             { CMD_READ_BULK, 0x04, 36, 0 },
         };
         // Write initial packet
-        if(!usbsend(kb, data_pkt[0], 1))
+        if(!usbsend(kb, data_pkt[0], MSG_SIZE, 1))
             return -1;
         // Read colors
         for(int i = 1; i < 5; i++){
-            if(!usbrecv(kb, data_pkt[i],in_pkt[i - 1]))
+            if(!usbrecv(kb, data_pkt[i], MSG_SIZE, in_pkt[i - 1]))
                 return -1;
             if(memcmp(in_pkt[i - 1], data_pkt[i], 4)){
                 ckb_err("Bad input header");
