@@ -9,10 +9,12 @@
 #include "usb.h"
 #include "keymap_patch.h"
 #include <ckbnextconfig.h>
+#include "legacykb_proto.h"
 
 // Values taken from the official website
 // Mice not in the list default to 12000 in the GUI
-dpi_list mouse_dpi_list[] = {
+const dpi_list mouse_dpi_list[] = {
+    { P_M55_RGB_PRO, 12400 },
     { P_M65, 8200 },
     { P_M65_PRO, 12000 },
     { P_M65_RGB_ELITE, 18000 },
@@ -27,17 +29,29 @@ dpi_list mouse_dpi_list[] = {
     { P_SABRE_O2, 6400 },
     { P_HARPOON, 6000 },
     { P_HARPOON_PRO, 12000 },
+    { P_KATAR_PRO_XT, 18000 },
     { P_KATAR, 8000 },
+    { P_KATAR_PRO, 12400 },
     { P_IRONCLAW, 18000 },
     { P_NIGHTSWORD, 18000},
+    { P_IRONCLAW_W_U, 18000 },
+    { P_HARPOON_WL_U, 10000 },
+    { P_GLAIVE_PRO, 16000 },
     { 0, 0 }, // Keep last and do not remove
 };
 
 // WARNING: macOS doesn't support idVendor != 0x1b1c at the moment
-device_desc models[] = {
+const device_desc models[] = {
     // Keyboards
     { V_CORSAIR, P_K55, },
+    { V_CORSAIR, P_K60_PRO_RGB, },
+    { V_CORSAIR, P_K60_PRO_RGB_LP, },
+    { V_CORSAIR, P_K60_PRO_RGB_SE, },
     { V_CORSAIR, P_K63_NRGB, },
+    { V_CORSAIR, P_K63_NRGB_WL, },
+    { V_CORSAIR, P_K63_NRGB_WL2, },
+    { V_CORSAIR, P_K63_NRGB_WL3, },
+    { V_CORSAIR, P_K63_NRGB_WL4, },
     { V_CORSAIR, P_K65, },
     { V_CORSAIR, P_K65_LEGACY, },
     { V_CORSAIR, P_K65_LUX, },
@@ -62,7 +76,11 @@ device_desc models[] = {
     { V_CORSAIR, P_STRAFE_NRGB, },
     { V_CORSAIR, P_STRAFE_NRGB_2, },
     { V_CORSAIR, P_STRAFE_MK2, },
+    { V_CORSAIR, P_K95_PLATINUM_XT, },
+    { V_CORSAIR, P_K57_D, },
+    { V_CORSAIR, P_K57_U, },
     // Mice
+    { V_CORSAIR, P_M55_RGB_PRO, },
     { V_CORSAIR, P_M65, },
     { V_CORSAIR, P_M65_PRO, },
     { V_CORSAIR, P_M65_RGB_ELITE, },
@@ -78,15 +96,28 @@ device_desc models[] = {
     { V_CORSAIR, P_HARPOON, },
     { V_CORSAIR, P_HARPOON_PRO, },
     { V_CORSAIR, P_KATAR, },
+    { V_CORSAIR, P_KATAR_PRO, },
+    { V_CORSAIR, P_KATAR_PRO_XT, },
     { V_CORSAIR, P_IRONCLAW, },
     { V_CORSAIR, P_NIGHTSWORD, },
+    { V_CORSAIR, P_DARK_CORE, },
+    { V_CORSAIR, P_DARK_CORE_SE, },
+    { V_CORSAIR, P_DARK_CORE_WL, },
+    { V_CORSAIR, P_DARK_CORE_SE_WL, },
+    { V_CORSAIR, P_DARK_CORE_RGB_PRO_SE, },
+    { V_CORSAIR, P_DARK_CORE_RGB_PRO_SE_WL, },
+    { V_CORSAIR, P_IRONCLAW_W_U, },
+    { V_CORSAIR, P_IRONCLAW_W_D, },
+    { V_CORSAIR, P_HARPOON_WL_U, },
+    { V_CORSAIR, P_HARPOON_WL_D, },
+    { V_CORSAIR, P_GLAIVE_PRO, },
     // Mousepads
     { V_CORSAIR, P_POLARIS, },
     // Headset stands
     { V_CORSAIR, P_ST100, },
 };
 
-size_t N_MODELS = sizeof(models) / sizeof(device_desc);
+const size_t N_MODELS = sizeof(models) / sizeof(device_desc);
 
 device_desc ignored_devices[DEV_MAX] = {0};
 
@@ -108,6 +139,9 @@ _Atomic int reset_stop = 0;
 /// \n Have a look at \a main() in main.c for details.
 int features_mask = -1;
 
+// Enable experimental support for not fully supported devices
+int enable_experimental = 0;
+
 /// brief .
 ///
 /// vendor_str returns "corsair" if the given \a vendor argument is equal to \a V_CORSAIR \c (0x1bc)
@@ -142,9 +176,11 @@ const char* vendor_str(ushort vendor){
 /// product_str() needs the \a product \a ID
 ///
 const char* product_str(ushort product){
-    if(product == P_K95 || product == P_K95_LEGACY)
+    if(product == P_K95_LEGACY)
+        return "k95l";
+    if(product == P_K95)
         return "k95";
-    if(product == P_K95_PLATINUM)
+    if(product == P_K95_PLATINUM || product == P_K95_PLATINUM_XT)
         return "k95p";
     if(product == P_K70 || product == P_K70_LEGACY || product == P_K70_LUX || product == P_K70_LUX_NRGB || product == P_K70_RFIRE || product == P_K70_RFIRE_NRGB)
         return "k70";
@@ -158,6 +194,12 @@ const char* product_str(ushort product){
         return "k66";
     if(product == P_K63_NRGB)
         return "k63";
+    if(product == P_K63_NRGB_WL || product == P_K63_NRGB_WL2 || product == P_K63_NRGB_WL3 || product == P_K63_NRGB_WL4)
+        return "k63_wireless";
+    if(product == P_K60_PRO_RGB || product == P_K60_PRO_RGB_LP || product == P_K60_PRO_RGB_SE)
+        return "k60";
+    if(product == P_K57_U || product == P_K57_D)
+        return "k57_wireless";
     if(product == P_K55)
         return "k55";
     if(product == P_STRAFE || product == P_STRAFE_NRGB || product == P_STRAFE_NRGB_2)
@@ -166,6 +208,8 @@ const char* product_str(ushort product){
         return "strafe_mk2";
     if(product == P_M95)
         return "m95";
+    if(product == P_M55_RGB_PRO)
+        return "m55";
     if(product == P_M65 || product == P_M65_PRO)
         return "m65";
     if(product == P_M65_RGB_ELITE)
@@ -176,16 +220,25 @@ const char* product_str(ushort product){
         return "scimitar";
     if(product == P_HARPOON || product == P_HARPOON_PRO)
         return "harpoon";
+    // Katar PRO XT and Katar PRO works the same, so we just do that
+    if(product == P_KATAR_PRO_XT || product == P_KATAR_PRO)
+        return "katarproxt";
     if(product == P_GLAIVE)
         return "glaive";
+    if(product == P_GLAIVE_PRO)
+        return "glaivepro";
     if(product == P_KATAR)
         return "katar";
     if(product == P_IRONCLAW)
         return "ironclaw";
     if(product == P_NIGHTSWORD)
         return "nightsword";
+    if(product == P_IRONCLAW_W_U || product == P_IRONCLAW_W_D || product == P_DARK_CORE_RGB_PRO_SE || product == P_DARK_CORE_RGB_PRO_SE_WL || product == P_HARPOON_WL_U || product == P_HARPOON_WL_D)
+        return "ironclaw_wireless";
     if(product == P_POLARIS)
         return "polaris";
+    if(product == P_DARK_CORE || product == P_DARK_CORE_WL || product == P_DARK_CORE_SE || product == P_DARK_CORE_SE_WL)
+        return "darkcore";
     if(product == P_ST100)
         return "st100";
     return "";
@@ -205,11 +258,21 @@ const char* product_str(ushort product){
 ///
 /// \todo Is the last point really a good decision and always correct?
 ///
-static const devcmd* get_vtable(ushort vendor, ushort product){
-    // return IS_MOUSE(vendor, product) ? &vtable_mouse : !IS_LEGACY(vendor, product) ? &vtable_keyboard : &vtable_keyboard_nonrgb;
-    if(IS_MOUSE(vendor, product)) {
+static const devcmd* get_vtable(usbdevice* kb){
+    ushort vendor = kb->vendor;
+    ushort product = kb->product;
+    if(kb->protocol == PROTO_BRAGI) {
+        if(IS_DONGLE(kb))
+            return &vtable_bragi_dongle;
+        else if(IS_MOUSE(vendor, product))
+            return &vtable_bragi_mouse;
+        else
+            return &vtable_bragi_keyboard;
+    } else if(IS_MOUSE(vendor, product)) {
         if(IS_LEGACY(vendor, product))
             return &vtable_mouse_legacy;
+        else if(IS_WIRELESS(vendor, product))
+            return &vtable_mouse_wireless;
         else
             return &vtable_mouse;
     } else if(IS_MOUSEPAD(vendor, product) || product == P_ST100) {
@@ -217,6 +280,8 @@ static const devcmd* get_vtable(ushort vendor, ushort product){
     } else {
         if(IS_LEGACY(vendor, product))
             return &vtable_keyboard_legacy;
+        else if(IS_WIRELESS(vendor, product))
+            return &vtable_keyboard_wireless;
         else
             return &vtable_keyboard;
     }
@@ -265,24 +330,23 @@ static void* devmain(usbdevice* kb){
     while(1){
         ///
         /// If the reading via readlines() is successful (we might have read multiple lines),
-        /// the interpretation is done by readcmd() iff the connection to the device is still available
-        /// (checked via IS_CONNECTED(kb)).
+        /// the interpretation is done by readcmd() if the connection to the device is still available
         /// This is true if the kb-structure has a handle and an event pointer both != Null).
         /// If not, the loop is left (the first exit point).
         queued_mutex_unlock(dmutex(kb));
         // Read from FIFO
         const char* line;
         int lines = readlines(kbfifo, linectx, &line);
+        wait_until_suspend_processed();
         queued_mutex_lock(dmutex(kb));
         // End thread when the handle is removed
-        if(!IS_CONNECTED(kb))
+        if(kb->status == DEV_STATUS_DISCONNECTING || kb->status == DEV_STATUS_DISCONNECTED)
             break;
         ///
         /// if nothing is in the line buffer (some magic interrupt?),
         /// continue in the endless while without any reaction.
         if(lines){
             /// \todo readcmd() gets a \b line, not \b lines. Have a look on that later.
-            /// \n Is the condition IS_CONNECTED valid? What functions change the condititon for the macro?
             if(readcmd(kb, line)){
                 ///
                 /// If interpretation and communication with the usb device got errors,
@@ -345,21 +409,39 @@ static void* _setupusb(void* context){
     ///
     usbdevice* kb = context;
     queued_mutex_lock(dmutex(kb));
+    if(USES_BRAGI(kb->vendor, kb->product))
+        kb->protocol = PROTO_BRAGI;
+
+    if(USES_BRAGI_JUMBO(kb->vendor, kb->product))
+        kb->out_ep_packet_size = BRAGI_JUMBO_SIZE;
+    else
+        kb->out_ep_packet_size = MSG_SIZE;
+
     queued_mutex_lock(imutex(kb));
     // Set standard fields
     ushort vendor = kb->vendor, product = kb->product;
-    const devcmd* vt = kb->vtable = get_vtable(vendor, product);
-    kb->features = (IS_LEGACY(vendor, product) ? FEAT_STD_LEGACY : FEAT_STD_RGB) & features_mask;
+    const devcmd* vt = get_vtable(kb);
+    memcpy(&kb->vtable, vt, sizeof(devcmd));
+    vt = &kb->vtable;
+
+    if(!(IS_DONGLE(kb) && kb->protocol == PROTO_BRAGI))
+        kb->features = (IS_LEGACY(vendor, product) ? FEAT_STD_LEGACY : FEAT_STD_RGB) & features_mask;
     if(SUPPORTS_ADJRATE(kb))
         kb->features |= FEAT_ADJRATE;
     if(IS_MONOCHROME(vendor, product))
         kb->features |= FEAT_MONOCHROME;
-    kb->usbdelay = USB_DELAY_DEFAULT;
+    if(IS_DONGLE(kb))
+        kb->features |= (FEAT_DONGLE | FEAT_FWVERSION);
+    if(IS_WIRELESS_DEV(kb)){
+        kb->features |= FEAT_WIRELESS;
+        if((kb->protocol == PROTO_BRAGI && !IS_DONGLE(kb)) || kb->protocol != PROTO_BRAGI)
+            kb->features |= FEAT_BATTERY;
+    }
+    // Disable FWUPDATE for all bragi devices
+    if(kb->protocol == PROTO_BRAGI)
+        kb->features &= ~FEAT_FWUPDATE;
 
-    /// Allocate memory for the os_usbrecv() buffer
-    kb->interruptbuf = malloc(MSG_SIZE * sizeof(uchar));
-    if(!kb->interruptbuf)
-        ckb_fatal("Error allocating memory for usb_recv() %s", strerror(errno));
+    kb->usbdelay = USB_DELAY_DEFAULT;
 
     // Check if the device needs a patched keymap, and if so patch it.
     patchkeys(kb);
@@ -375,8 +457,9 @@ static void* _setupusb(void* context){
     /// As a result, some parameters should be set in kb (name, serial, fwversion, epcount = number of usb endpoints),
     /// and all endpoints should be claimed with usbclaim().
     /// Claiming is the only point where os_setupusb() can produce an error (-1, otherwise 0).
-    if(os_setupusb(kb))
-        goto fail;
+    if(!kb->parent)
+        if(os_setupusb(kb))
+            goto fail;
 
     ///
     /// - The following two statements deal with possible errors when setting the kb values in the current routine:
@@ -389,6 +472,9 @@ static void* _setupusb(void* context){
     if(!kb->name[0])
         snprintf(kb->name, KB_NAME_LEN, "%s %s", vendor_str(kb->vendor), product_str(kb->product));
 
+    // Ask the protocol handler to set the endpoints required for the main input thread
+    kb->vtable.fill_input_eps(kb);
+
     // Set up an input device for key events
     ///
     /// - Then the user level input subsystem is activated via os_openinput().
@@ -402,26 +488,28 @@ static void* _setupusb(void* context){
     ///
     /// - The next action is to create a separate thread, which gets as parameter kb and starts with os_inputmain().
     /// The thread is immediately detached so that it can return its resource completely independently if it should terminate.
-    if(pthread_create(&kb->inputthread, 0, os_inputmain, kb))
-        goto fail;
+    if(!kb->parent){
+        if(pthread_create(&kb->inputthread, 0, os_inputmain, kb))
+            goto fail;
 
 #ifndef OS_MAC
-    // name this thread externally if not on mac,
-    // on mac `pthread_setname_np` is only naming the current thread
-    char inputthread_name[THREAD_NAME_MAX] = "ckbX input";
-    inputthread_name[3] = INDEX_OF(kb, keyboard) + '0';
-    pthread_setname_np(kb->inputthread, inputthread_name);
+        // name this thread externally if not on mac,
+        // on mac `pthread_setname_np` is only naming the current thread
+        char inputthread_name[THREAD_NAME_MAX] = "ckbX input";
+        inputthread_name[3] = INDEX_OF(kb, keyboard) + '0';
+        pthread_setname_np(kb->inputthread, inputthread_name);
 #endif // OS_MAC
-    pthread_detach(kb->inputthread);
+        pthread_detach(kb->inputthread);
 
-    ///
-    /// - The same happens with os_setupindicators(),
-    /// which initially initializes all LED variables in kb to off and then starts the _ledthread() thread
-    /// with kb as parameter and then detaches it.
-    /// Here again only the generation of the thread can fail.
-    if(!(IS_MOUSE_DEV(kb) || IS_MOUSEPAD_DEV(kb))) {
-        if(os_setupindicators(kb))
-            goto fail;
+        ///
+        /// - The same happens with os_setupindicators(),
+        /// which initially initializes all LED variables in kb to off and then starts the _ledthread() thread
+        /// with kb as parameter and then detaches it.
+        /// Here again only the generation of the thread can fail.
+        if(!(IS_MOUSE_DEV(kb) || IS_MOUSEPAD_DEV(kb))) {
+            if(os_setupindicators(kb))
+                goto fail;
+        }
     }
 
     // Set up device
@@ -494,9 +582,9 @@ static void* _setupusb(void* context){
     // Finished. Enter main loop
     int index = INDEX_OF(kb, keyboard);
     ckb_info("Setup finished for %s%d", devpath, index);
-    queued_mutex_unlock(dmutex(kb));
+    kb->status = DEV_STATUS_CONNECTED;
     updateconnected(kb);
-    queued_mutex_lock(dmutex(kb));
+
     ///
     /// devmain()'s return value is returned by _setupusb() when we terminate.
     return devmain(kb);
@@ -551,7 +639,9 @@ void setupusb(usbdevice* kb){
 int revertusb(usbdevice* kb){
     if(NEEDS_FW_UPDATE(kb))
         return 0;
-    if(!HAS_FEATURES(kb, FEAT_RGB)){
+
+    // FIXME: This should be moved to a device-specific function
+    if(IS_LEGACY_DEV(kb)){
         nk95cmd(kb, NK95_HWON);
         return 0;
     }
@@ -575,9 +665,9 @@ int _resetusb(usbdevice* kb, const char* file, int line){
         return res;
     DELAY_LONG(kb);
     // Re-initialize the device
-    if(kb->vtable->start(kb, kb->active) != 0)
+    if(kb->vtable.start(kb, kb->active) != 0)
         return -1;
-    if(kb->vtable->updatergb(kb, 1) != 0)
+    if(kb->vtable.updatergb(kb, 1) != 0)
         return -1;
     return 0;
 }
@@ -628,59 +718,15 @@ int usb_tryreset(usbdevice* kb){
 ///
 extern int hwload_mode;
 
-/// \brief .
-///
-/// \todo A lot of different conditions are combined in this code. Don't think, it is good in every combination...
-///
-/// The main task of _usbsend () is to transfer the complete logical message from the buffer beginning with \a messages to <b>count * MSG_SIZE</b>.
-/// \n According to usb 2.0 specification, a USB transmits a maximum of 64 byte user data packets.
-/// For the transmission of longer messages we need a segmentation.
-/// And that is exactly what happens here.
-///
-/// The message is given one by one to os_usbsend() in MSG_SIZE (= 64) byte large bites.
-/// \attention This means that the buffer given as argument must be n * MSG_SIZE Byte long.
-///
-/// An essential constant parameter which is relevant for os_usbsend() only is is_recv = 0, which means sending.
-///
-/// Now it gets a little complicated again:
-/// - If os_usbsend() returns 0, only zero bytes could be sent in one of the packets,
-/// or it was an error (-1 from the systemcall), but not a timeout.
-/// How many Bytes were sent in total from earlier calls does not seem to matter,
-/// _usbsend() returns a total of 0.
-/// - Returns os_usbsend() -1, first check if \b reset_stop is set globally
-/// or (incomprehensible) hwload_mode is not set to "always".
-/// In either case, _usbsend() returns 0,
-/// otherwise it is assumed to be a temporary transfer error and it simply retransmits the physical packet after a long delay.
-/// - If the return value of os_usbsend() was neither 0 nor -1,
-/// it specifies the numer of bytes transferred.
-/// \n Here is an information hiding conflict with os_usbsend() (at least in the Linux version):
-/// \n If os_usbsend() can not transfer the entire packet,
-/// errors are thrown and the number of bytes sent is returned.
-/// _usbsend() interprets this as well
-/// and remembers the total number of bytes transferred in the local variable \b total_sent.
-/// Subsequently, however, transmission is continued with the next complete MSG_SIZE block
-/// and not with the first of the possibly missing bytes.
-/// \todo Check whether this is the same in the macOS variant. It is not dramatic, but if errors occur, it can certainly irritate the devices completely if they receive incomplete data streams. Do we have errors with the messages "Wrote YY bytes (expected 64)" in the system logs? If not, we do not need to look any further.
-///
-/// When the last packet is transferred,
-/// _usbsend() returns the effectively counted set of bytes (from \b total_sent).
-/// This at least gives the caller the opportunity to check whether something has been lost in the middle.
-///
-/// A bit strange is the structure of the program:
-/// Handling the \b count MSG_SIZE blocks to be transferred is done
-/// in the outer for (...) loop.
-/// Repeating the transfer with a treatable error is managed by the inner while(1) loop.
-/// \n This must be considered when reading the code;
-/// The "break" on successful block transfer leaves the inner while, not the for (...).
-///
-int _usbsend(usbdevice* kb, const uchar* messages, int count, const char* file, int line){
+// Wrapper around the vtable write() function for error handling and recovery
+int _usbsend(usbdevice* kb, void* messages, size_t msg_len, int count, const char* file, int line){
     int total_sent = 0;
     for(int i = 0; i < count; i++){
         // Send each message via the OS function
         while(1){
             DELAY_SHORT(kb);
             queued_mutex_lock(mmutex(kb)); ///< Synchonization between macro and color information
-            int res = os_usbsend(kb, messages + i * MSG_SIZE, 0, file, line);
+            int res = kb->vtable.write(kb, messages + i * msg_len, msg_len, 0, file, line);
             queued_mutex_unlock(mmutex(kb));
             if(res == 0)
                 return 0;
@@ -698,76 +744,16 @@ int _usbsend(usbdevice* kb, const uchar* messages, int count, const char* file,
     return total_sent;
 }
 
-int _usbsend_control(usbdevice* kb, uchar* data, ushort len, uchar bRequest, ushort wValue, ushort wIndex, const char* file, int line){
-    while(1){
-        DELAY_SHORT(kb);
-        queued_mutex_lock(mmutex(kb)); ///< Synchonization between macro and color information
-        int res = os_usbsend_control(kb, data, len, bRequest, wValue, wIndex, file, line);
-        queued_mutex_unlock(mmutex(kb));
+// Wrapper around the vtable write() and read() functions for error handling and recovery
+int _usbrecv(usbdevice* kb, void* out_msg, size_t msg_len, uchar* in_msg, const char* file, int line){
+    // For now assume that msg_len is for both out_msg and in_msg
 
-        if(res != -1)
-            return res;
-
-        // Stop immediately if the program is shutting down or hardware load is set to tryonce
-        if(reset_stop || hwload_mode != 2)
-            return 0;
-
-        // Retry as long as the result is temporary failure
-        DELAY_LONG(kb);
-    }
-}
-
-/// \brief .
-///
-/// To fully understand this, you need to know about usb:
-/// All control is at the usb host (the CPU).
-/// If the device wants to communicate something to the host,
-/// it must wait for the host to ask.
-/// The usb protocol defines the cycles and periods in which actions are to be taken.
-///
-/// So in order to receive a data packet from the device,
-/// the host must first send a send request.
-/// \n This is done by _usbrecv() in the first block
-/// by sending the MSG_SIZE large data block from \b out_msg via os_usbsend() as it is a machine depending implementation.
-/// The usb target device is as always determined over kb.
-///
-/// For os_usbsend() to know that it is a receive request,
-/// the \b is_recv parameter is set to true (1).
-/// With this, os_usbsend () generates a control package for the hardware, not a data packet.
-///
-/// If sending of the control package is not successful,
-/// a maximum of 5 times the transmission is repeated (including the first attempt).
-/// If a non-cancelable error is signaled or the drive is stopped via reset_stop,
-/// _usbrecv() immediately returns 0.
-///
-/// After this, the function waits for the requested response from the device using os_usbrecv ().
-///
-/// os_usbrecv() returns 0, -1 or something else.
-/// \n Zero signals a serious error which is not treatable and _usbrecv() also returns 0.
-/// \n -1 means that it is a treatable error - a timeout for example -
-/// and therefore the next transfer attempt is started after a long pause (DELAY_LONG)
-/// if not reset_stop or the wrong hwload_mode require a termination with a return value of 0.
-///
-/// After 5 attempts, _usbrecv () returns and returns 0 as well as an error message.
-///
-/// When data is received, the number of received bytes is returned.
-/// This should always be MSG_SIZE,
-/// but os_usbrecv() can also return less.
-/// It should not be more,
-/// because then there would be an unhandled buffer overflow,
-/// but it could be less.
-/// This would be signaled in os_usbrecv () with a message.
-///
-/// The buffers behind \b out_msg and \b in_msg are MSG_SIZE at least (currently 64 Bytes).
-/// More is ok but useless, less brings unpredictable behavior.
-///
-int _usbrecv(usbdevice* kb, const uchar* out_msg, uchar* in_msg, const char* file, int line){
     // Try a maximum of 5 times
     for (int try = 0; try < 5; try++) {
         // Send the output message
         queued_mutex_lock(mmutex(kb)); ///< Synchonization between macro and color information
         DELAY_SHORT(kb);
-        int res = os_usbsend(kb, out_msg, 1, file, line);
+        int res = kb->vtable.write(kb, out_msg, msg_len, 1, file, line);
         queued_mutex_unlock(mmutex(kb));
         if (res == 0)
             return 0;
@@ -779,16 +765,16 @@ int _usbrecv(usbdevice* kb, const uchar* out_msg, uchar* in_msg, const char* fil
             continue;
         }
         // Wait for the response
-        if(!(kb->fwversion >= 0x120 || IS_V2_OVERRIDE(kb)))
+        if(!(kb->fwversion >= 0x120 || IS_V2_OVERRIDE(kb)) && kb->protocol != PROTO_BRAGI)
             DELAY_MEDIUM(kb);
-        res = os_usbrecv(kb, in_msg, file, line);
+        res = kb->vtable.read(kb, in_msg, msg_len, 0, file, line);
         if(res == 0)
             return 0;
         else if(res != -1)
             return res;
         if(reset_stop || hwload_mode != 2)
             return 0;
-        if(!(kb->fwversion >= 0x120 || IS_V2_OVERRIDE(kb)))
+        if(!(kb->fwversion >= 0x120 || IS_V2_OVERRIDE(kb)) && kb->protocol != PROTO_BRAGI)
             DELAY_LONG(kb);
     }
     // Give up
@@ -840,73 +826,87 @@ int _usbrecv(usbdevice* kb, const uchar* out_msg, uchar* in_msg, const char* fil
 /// and closeusb() always returns zero (success).
 ///
 int closeusb(usbdevice* kb){
-    queued_mutex_lock(imutex(kb));
-    if(kb->handle){
-        int index = INDEX_OF(kb, keyboard);
-        ckb_info("Disconnecting %s%d", devpath, index);
-        os_inputclose(kb);
-        queued_mutex_unlock(imutex(kb));
-        queued_mutex_unlock(dmutex(kb));
-        updateconnected(kb);
-        queued_mutex_lock(dmutex(kb));
-        queued_mutex_lock(imutex(kb));
-        // Close USB device
-        os_closeusb(kb);
-    } else {
-        queued_mutex_unlock(imutex(kb));
-        queued_mutex_unlock(dmutex(kb));
-        updateconnected(kb);
-        queued_mutex_lock(dmutex(kb));
-        queued_mutex_lock(imutex(kb));
+    kb->status = DEV_STATUS_DISCONNECTING;
+    const int index = INDEX_OF(kb, keyboard);
+    ckb_info("Disconnecting %s%d", devpath, index);
+    updateconnected(kb);
+
+    // If the device has children, disconnect them first
+    pthread_mutex_lock(cmutex(kb));
+    for(int i = 0; i < MAX_CHILDREN; i++){
+        if(!kb->children[i])
+            continue;
+        queued_mutex_lock(dmutex(kb->children[i]));
+        closeusb(kb->children[i]);
+        queued_mutex_unlock(dmutex(kb->children[i]));
+        kb->children[i] = NULL;
     }
+
+    queued_mutex_lock(imutex(kb));
+    os_inputclose(kb);
+
+    // Close USB device
+    os_closeusb(kb);
     rmdevpath(kb);
 
-    // Wait for thread to close
-    queued_mutex_unlock(imutex(kb));
-    queued_mutex_unlock(dmutex(kb));
+    // Wait for threads to close
+    if(kb->pollthread || kb->thread){
+        queued_mutex_unlock(imutex(kb));
+        queued_mutex_unlock(dmutex(kb));
 
-    if(pthread_equal(pthread_self(), kb->thread)){
+        // Shut down the device polling thread
+        if(kb->pollthread){
+            // Unlock dmutex to allow pollthread to quit
+            pthread_kill(*kb->pollthread, SIGUSR2);
+            pthread_join(*kb->pollthread, NULL);
+            free(kb->pollthread);
+            kb->pollthread = NULL;
+        }
+
+        if(kb->thread){
+            if(pthread_equal(pthread_self(), kb->thread)){
 #ifdef DEBUG_MUTEX
-        ckb_info("Attempted to pthread_join() self. Detaching instead.");
+                ckb_info("Attempted to pthread_join() self. Detaching instead.");
 #endif
-        int detachres = pthread_detach(kb->thread);
-        if(detachres)
-            ckb_err("pthread_detach() returned %s (%d)", strerror(detachres), detachres);
-    } else {
+                int detachres = pthread_detach(kb->thread);
+                if(detachres)
+                    ckb_err("pthread_detach() returned %s (%d)", strerror(detachres), detachres);
+            } else {
 #ifdef DEBUG_MUTEX
-        ckb_info("Joining thread 0x%lx for ckb%d by thread 0x%lx", kb->thread, INDEX_OF(kb, keyboard), pthread_self());
+                ckb_info("Joining thread 0x%lx for ckb%d by thread 0x%lx", kb->thread, INDEX_OF(kb, keyboard), pthread_self());
 #endif
-        int joinres = pthread_join(kb->thread, NULL);
-        if(joinres)
-            ckb_err("pthread_join() returned %s (%d)", strerror(joinres), joinres);
+                int joinres = pthread_join(kb->thread, NULL);
+                if(joinres)
+                    ckb_err("pthread_join() returned %s (%d)", strerror(joinres), joinres);
+            }
+        }
+        queued_mutex_lock(dmutex(kb));
+        queued_mutex_lock(imutex(kb));
     }
-    queued_mutex_lock(dmutex(kb));
 
-    // Free the device-specific keymap
-    free(kb->keymap);
-
-    // Free the memory used for the os_usbrecv() buffer
-    free(kb->interruptbuf);
-
-    // Delete the profile and the control path
-    if(!kb->vtable)
-        return 0;
-    kb->vtable->freeprofile(kb);
-    queued_mutex_lock(imutex(kb));
+    // Delete the profile
+    if(kb->vtable.freeprofile)
+        kb->vtable.freeprofile(kb);
+    // This implicitly sets the status to STATUS_DISCONNECTED
     memset(kb, 0, sizeof(usbdevice));
+    pthread_mutex_unlock(cmutex(kb));
     queued_mutex_unlock(imutex(kb));
     return 0;
 }
 
 /// Formats and writes the current urb buffer to the console
-void print_urb_buffer(const char* prefix, const unsigned char* buffer, int actual_length, const char* file, int line, const char* function, int devnum){
-    char converted[actual_length * 3 + 1];
+void print_urb_buffer(const char* prefix, const unsigned char* buffer, int actual_length, const char* file, int line, const char* function, int devnum, const uchar ep){
+    char* converted = malloc(actual_length * 3 + 1);
     for(int i = 0; i < actual_length; i++)
-        sprintf(&converted[i * 3], "%02x ", buffer[i]);
+        sprintf(converted + i * 3, "%02x ", buffer[i]);
+    char ep_str[6] = {0};
+    if(ep)
+        snprintf(ep_str, sizeof(ep_str), " (%02x)", ep);
     if(line == 0)
-        ckb_info("ckb%i %s %s", devnum, prefix, converted);
+        ckb_info("ckb%d %s%s %s", devnum, prefix, ep_str, converted);
     else
-        ckb_info("ckb%i %s (via %s:%d) %s %s", devnum, function, file, line, prefix, converted);
+        ckb_info("ckb%d %s (via %s:%d) %s%s %s", devnum, function, file, line, prefix, ep_str, converted);
+    free(converted);
 }
 
 #define TRY_WITH_RESET_NOFREE(action, ret)       \
@@ -926,11 +926,11 @@ void reactivate_devices(){
     for(int i = 1; i < DEV_MAX; i++){
         usbdevice* kb = keyboard + i;
         queued_mutex_lock(dmutex(kb));
-        if(IS_CONNECTED(kb) && !NEEDS_FW_UPDATE(kb) && kb->active){
+        if(kb->status == DEV_STATUS_CONNECTED && !NEEDS_FW_UPDATE(kb) && kb->active){
             int ret;
             // If the device was active, mark it as disabled and re-enable it
             kb->active = 0;
-            const devcmd* vt = kb->vtable;
+            const devcmd* vt = &kb->vtable;
             TRY_WITH_RESET_NOFREE(vt->active(kb, 0, 0, 0, 0), ret);
             TRY_WITH_RESET_NOFREE(vt->updatergb(kb, 1), ret);
             TRY_WITH_RESET_NOFREE(vt->updatedpi(kb, 1), ret);
