@@ -9,279 +9,72 @@
 
 // usb.c
 extern _Atomic int reset_stop;
+extern int enable_experimental;
 
 /// \details
 /// \brief all open usb devices have their system path names here in this array.
-
 static char kbsyspath[DEV_MAX][FILENAME_MAX];
 
-int os_usbsend_control(usbdevice* kb, uchar* data, ushort len, uchar bRequest, ushort wValue, ushort wIndex, const char* file, int line) {
+// USB IO functions
+int os_usb_control(usbdevice* kb, ctrltransfer* transfer, const char* file, int line) {
 #ifdef DEBUG_USB_SEND
-    int ckb = INDEX_OF(kb, keyboard);
-    ckb_info("ckb%d Control (%s:%d): bmRequestType: 0x%02hhx, bRequest: %hhu, wValue: 0x%04hx, wIndex: %04hx, wLength: %hu", ckb, file, line, 0x40, bRequest, wValue, wIndex, len);
-    if(len)
-        print_urb_buffer("Control buffer:", data, len, file, line, __func__, ckb);
+    const int ckb = INDEX_OF(kb, keyboard);
+    ckb_info("ckb%d Control (%s:%d): bmRequestType: 0x%02hhx, bRequest: %hhu, wValue: 0x%04hx, wIndex: %04hx, wLength: %hu", ckb, file, line, transfer->bRequestType, transfer->bRequest, transfer->wValue, transfer->wIndex, transfer->wLength);
+    if(transfer->wLength)
+        print_urb_buffer("Control buffer:", transfer->data, transfer->wLength, file, line, __func__, ckb, 0);
 #endif
 
-    struct usbdevfs_ctrltransfer transfer = { 0x40, bRequest, wValue, wIndex, len, 5000, data };
-    int res = ioctl(kb->handle - 1, USBDEVFS_CONTROL, &transfer);
+    const int res = ioctl(kb->handle - 1, USBDEVFS_CONTROL, transfer);
     if (res == -1){
         int ioctlerrno = errno;
         ckb_err_fn(" %s, res = 0x%x", file, line, strerror(ioctlerrno), res);
+#ifndef NDEBUG
+        if (ioctlerrno == EBUSY) {
+            if ((transfer->bRequestType & 0x1f) == 1) {
+                struct usbdevfs_getdriver query = {
+                    .interface = transfer->wIndex,
+                };
+                if (!ioctl(kb->handle - 1, USBDEVFS_GETDRIVER, &query)) {
+                    ckb_info("Directed at interface %d, which is claimed by %s", transfer->wIndex, query.driver);
+                } else {
+                    ckb_info("Directed at interface %d, GETDRIVER error %s", transfer->wIndex, strerror(errno));
+                }
+            } else {
+                ckb_info("Directed at endpoint (called from where?)");
+            }
+        }
+#endif
         if(ioctlerrno == ETIMEDOUT)
             return -1;
         else
             return 0;
 
-    } else if (res != len)
+    } else if (res != transfer->wLength)
         ckb_warn_fn("Wrote %d bytes (expected %d)", file, line, res, MSG_SIZE);
 
     return res;
 }
 
-////
-/// \brief os_usbsend sends a data packet (MSG_SIZE = 64) Bytes long
-///
-/// os_usbsend has two functions:
-/// - if is_recv == false, it tries to send a given MSG_SIZE buffer via the usb interface given with kb.
-/// - otherwise a request is sent via the usb device to initiate the receiving of a message from the remote device.
-///
-/// The functionality for sending distinguishes two cases,
-/// depending on the version number of the firmware of the connected device:
-/// \n If the firmware is less or equal 1.2, the transmission is done via an ioctl().
-/// The ioctl() is given a struct usbdevfs_ctrltransfer, in which the relevant parameters are entered:
-///
-/// bRequestType | bRequest | wValue | EP | size | Timeout | data
-/// ------------ | -------- | ------ | -- | ---- | ------- | ----
-/// 0x21 | 0x09 | 0x0200 | endpoint / IF to be addressed from epcount-1 | MSG_SIZE | 5000 (=5ms) | the message buffer pointer
-/// Host to Device, Type=Class, Recipient=Interface | 9 = Send data? | specific | last or pre-last device # | 64 | 5000 | out_msg
-///
-/// \n The ioctl command is USBDEVFS_CONTROL.
-///
-/// The same constellation is used if the device is requested to send its data (is_recv = true).
-///
-/// For a more recent firmware and is_recv = false,
-/// the ioctl command USBDEVFS_CONTROL is not used
-/// (this tells the bus to enter the control mode),
-/// but the bulk method is used: USBDEVFS_BULK.
-/// This is astonishing, because all of the endpoints are type Interrupt, not bulk.
-///
-/// Anyhow, forthis purpose a different structure is used for the ioctl() (struct \b usbdevfs_bulktransfer)
-/// and this is also initialized differently:
-/// \n The length and timeout parameters are given the same values as above.
-/// The formal parameter out_msg is also passed as a buffer pointer.
-/// For the endpoints, the firmware version is differentiated again:
-/// \n For a firmware version between 1.3 and <2.0 endpoint 4 is used,
-/// otherwise (it can only be >=2.0) endpoint 3 is used.
-///
-/// \todo Since the handling of endpoints has already led to problems elsewhere, this implementation is extremely hardware-dependent and critical!
-/// \n Eg. the new keyboard K95PLATINUMRGB has a version number significantly less than 2.0 - will it run with this implementation?
-///
-/// The ioctl() - no matter what type -
-/// returns the number of bytes sent.
-/// Now comes the usual check:
-/// - If the return value is -1 AND the error is a timeout (ETIMEOUT),
-/// os_usbsend() will return -1 to indicate that it is probably a recoverable problem and a retry is recommended.
-/// - For another negative value or other error identifier OR 0 bytes sent, 0 is returned as a heavy error identifier.
-/// - In all other cases, the function returns the number of bytes sent.
-///
-/// If this is not the entire blocksize (MSG_SIZE bytes),
-/// an error message is issued on the standard error channel
-/// [warning "Wrote YY bytes (expected 64)"].
-///
-/// If DEBUG_USB_SEND is set during compilation,
-/// the number of bytes sent and their representation are logged to the error channel.
-///
-int os_usbsend(usbdevice* kb, const uchar* out_msg, int is_recv, const char* file, int line) {
-    int res;
-    if (kb->fwversion >= 0x120 || IS_V2_OVERRIDE(kb)){
-        // If we need to read a response, lock the interrupt mutex
-        if(is_recv)
-            if(pthread_mutex_lock(intmutex(kb)))
-                ckb_fatal("Error locking interrupt mutex in os_usbsend()");
-
-        struct usbdevfs_bulktransfer transfer = {0};
-        // All firmware versions for normal HID devices have the OUT endpoint at the end
-        // Devices with no input, such as the Polaris, have it at the start.
-        transfer.ep = (IS_SINGLE_EP(kb) ? 1 : kb->epcount);
-        transfer.len = MSG_SIZE;
-        transfer.timeout = 5000;
-        transfer.data = (void*)out_msg;
-        res = ioctl(kb->handle - 1, USBDEVFS_BULK, &transfer);
-    } else {
-        // Note, Ctrl Transfers require an index, not an endpoint, which is why kb->epcount - 1 works
-        struct usbdevfs_ctrltransfer transfer = { 0x21, 0x09, 0x0200, kb->epcount - 1, MSG_SIZE, 5000, (void*)out_msg };
-        res = ioctl(kb->handle - 1, USBDEVFS_CONTROL, &transfer);
-    }
-
+int os_usb_interrupt_out(usbdevice* kb, unsigned int ep, unsigned int len, uchar* data, const char* file, int line)
+{
+#ifdef DEBUG_USB_SEND
+    print_urb_buffer("Sending:", data, MSG_SIZE, file, line, __func__, INDEX_OF(kb, keyboard), (uchar)ep);
+#endif
+    const struct usbdevfs_bulktransfer transfer = { .ep = ep, .len = len, .timeout = 5000, .data = data, };
+    int res = ioctl(kb->handle - 1, USBDEVFS_BULK, &transfer);
     if (res <= 0){
         int ioctlerrno = errno;
         ckb_err_fn("%s, res = 0x%x", file, line, res ? strerror(ioctlerrno) : "No data written", res);
-        if(res == -1 && ioctlerrno == ETIMEDOUT){
-            if(is_recv)
-                pthread_mutex_unlock(intmutex(kb));
+        if(res == -1 && ioctlerrno == ETIMEDOUT)
             return -1;
-        } else {
-            if(is_recv)
-                pthread_mutex_unlock(intmutex(kb));
+        else
             return 0;
-        }
-    } else if (res != MSG_SIZE)
-        ckb_warn_fn("Wrote %d bytes (expected %d)", file, line, res, MSG_SIZE);
-#ifdef DEBUG_USB_SEND
-    print_urb_buffer("Sent:", out_msg, MSG_SIZE, file, line, __func__, INDEX_OF(kb, keyboard));
-#endif
-
+    } else if ((unsigned int)res != len) {
+        ckb_warn_fn("Wrote %d bytes (expected %d)", file, line, res, len);
+    }
     return res;
 }
 
-///
-/// \brief os_usbrecv receives a max MSGSIZE long buffer from usb device
-
-/// os_usbrecv does what its name says:
-///
-/// The comment at the beginning of the procedure
-/// causes the suspicion that the firmware versionspecific distinction
-/// is missing for receiving from usb endpoint 3 or 4.
-/// The commented code contains only the reception from EP4,
-/// but this may be wrong for a software version 2.0 or higher (see the code for os-usbsend ()).
-///
-/// \n So all the receiving is done via an ioctl() like in os_usbsend.
-/// The ioctl() is given a struct usbdevfs_ctrltransfer, in which the relevant parameters are entered:
-///
-/// bRequestType | bRequest | wValue | EP | size | Timeout | data
-/// ------------ | -------- | ------ | -- | ---- | ------- | ----
-/// 0xA1 | 0x01 | 0x0200 | endpoint to be addressed from epcount - 1 | MSG_SIZE | 5ms | the message buffer pointer
-/// Device to Host, Type=Class, Recipient=Interface | 1 = RECEIVE? | specific | Interface # | 64 | 5000 | in_msg
-///
-/// The ioctl() returns the number of bytes received.
-/// Here is the usual check again:
-/// - If the return value is -1 AND the error is a timeout (ETIMEOUT),
-/// os_usbrecv() will return -1 to indicate that it is probably a recoverable problem and a retry is recommended.
-/// - For another negative value or other error identifier OR 0 bytes are received, 0 is returned as an identifier for a heavy error.
-/// - In all other cases, the function returns the number of bytes received.
-///
-/// If this is not the entire blocksize (MSG_SIZE bytes),
-/// an error message is issued on the standard error channel
-/// [warning "Read YY bytes (expected 64)"].
-///
-int os_usbrecv(usbdevice* kb, uchar* in_msg, const char* file, int line){
-    int res;
-    if(kb->fwversion >= 0x120 || IS_V2_OVERRIDE(kb)){
-        // Wait for 2s
-        int condret = cond_nanosleep(intcond(kb), intmutex(kb), 2000000000);
-        if(condret != 0){
-            if(pthread_mutex_unlock(intmutex(kb)))
-                ckb_fatal("Error unlocking interrupt mutex in os_usbrecv()");
-            if(condret == ETIMEDOUT)
-                ckb_warn_fn("ckb%d: Timeout while waiting for response", file, line, INDEX_OF(kb, keyboard));
-            else
-                ckb_warn_fn("Interrupt cond error %i", file, line, condret);
-            return -1;
-        }
-        memcpy(in_msg, kb->interruptbuf, MSG_SIZE);
-        memset(kb->interruptbuf, 0, MSG_SIZE);
-        res = MSG_SIZE;
-        if(pthread_mutex_unlock(intmutex(kb)))
-            ckb_fatal("Error unlocking interrupt mutex in os_usbrecv()");
-    } else {
-        struct usbdevfs_ctrltransfer transfer = { 0xa1, 0x01, 0x0300, kb->epcount - 1, MSG_SIZE, 5000, in_msg };
-        res = ioctl(kb->handle - 1, USBDEVFS_CONTROL, &transfer);
-        if(res <= 0){
-            // This is done because ckb_err_fn can set errno itself
-            int ioctlerrno = errno;
-            ckb_err_fn("%s", file, line, res ? strerror(ioctlerrno) : "No data read");
-            if(res == -1 && ioctlerrno == ETIMEDOUT)
-                return -1;
-            else
-                return 0;
-        } else if(res != MSG_SIZE)
-            ckb_warn_fn("Read %d bytes (expected %d)", file, line, res, MSG_SIZE);
-    }
-
-#ifdef DEBUG_USB_RECV
-    print_urb_buffer("Recv:", in_msg, MSG_SIZE, file, line, __func__, INDEX_OF(kb, keyboard));
-#endif
-
-    return res;
-}
-
-///
-/// \brief _nk95cmd If we control a non RGB keyboard, set the keyboard via ioctl with usbdevfs_ctrltransfer
-///
-/// To send control packets to a non RGB non color K95 Keyboard,
-/// use this function. Normally it is called via the nk95cmd() macro.
-///
-/// If it is the wrong device for which the function is called, 0 is returned and nothing done.
-/// Otherwise a usbdevfs_ctrltransfer structure is filled and an USBDEVFS_CONTROL ioctl() called.
-///
-/// bRequestType | bRequest | wValue | EP | size | Timeout | data
-/// ------------ | -------- | ------ | -- | ---- | ------- | ----
-/// 0x40 | see table below to switch hardware-modus at Keyboard | wValue | device | MSG_SIZE | 5ms | the message buffer pointer
-/// Host to Device, Type=Vendor, Recipient=Device | bRequest parameter | given wValue Parameter | device 0 | 0 data to write | 5000 | null
-///
-/// If a 0 or a negative error number is returned by the ioctl, an error message is shown depending on the errno or "No data written" if retval was 0.
-/// In either case 1 is returned to indicate the error.
-/// If the ioctl returned a value > 0, 0 is returned to indicate no error.
-///
-/// Currently the following combinations for bRequest and wValue are used:
-/// Device | what it might to do | constant | bRequest | wValue
-/// ------ | ------------------- | -------- | -------- | ------
-/// non RGB Keyboard | set HW-modus on (leave the ckb driver) | HWON | 0x0002 | 0x0030
-/// non RGB Keyboard | set HW-modus off (initialize the ckb driver) | HWOFF | 0x0002 | 0x0001
-/// non RGB Keyboard | set light modus M1 in single-color keyboards | NK95_M1 | 0x0014 | 0x0001
-/// non RGB Keyboard | set light modus M2 in single-color keyboards | NK95_M2 | 0x0014 | 0x0002
-/// non RGB Keyboard | set light modus M3 in single-color keyboards | NK95_M3 | 0x0014 | 0x0003
-/// \see usb.h
-///
-int _nk95cmd(usbdevice* kb, uchar bRequest, ushort wValue, const char* file, int line){
-    if(kb->product != P_K95_LEGACY)
-        return 0;
-    struct usbdevfs_ctrltransfer transfer = { 0x40, bRequest, wValue, 0, 0, 5000, 0 };
-    int res = ioctl(kb->handle - 1, USBDEVFS_CONTROL, &transfer);
-    if(res <= 0){
-        ckb_err_fn("%s", file, line, res ? strerror(errno) : "No data written");
-        return 1;
-    }
-    return 0;
-}
-
-/// \brief .
-///
-/// \brief os_sendindicators update the indicators for the special keys (Numlock, Capslock and what else?)
-///
-/// Read the data from kb->ileds ans send them via ioctl() to the keyboard.
-///
-/// bRequestType | bRequest | wValue | EP | size | Timeout | data
-/// ------------ | -------- | ------ | -- | ---- | ------- | ----
-/// 0x21 | 0x09 | 0x0200 | Interface 0 | MSG_SIZE 1 Byte | timeout 0,5ms | the message buffer pointer
-/// Host to Device, Type=Class, Recipient=Interface (why not endpoint?) | 9 = SEND? | specific | 0 | 1 | 500 | struct* kb->ileds
-///
-/// \n The ioctl command is USBDEVFS_CONTROL.
-///
-void os_sendindicators(usbdevice* kb) {
-    static int countForReset = 0;
-    void *ileds;
-    ushort leds;
-    if(kb->fwversion >= 0x300 || IS_V3_OVERRIDE(kb)) {
-        leds = (kb->ileds << 8) | 0x0001;
-        ileds = &leds;
-    }
-    else {
-        ileds = &kb->ileds;
-    }
-    struct usbdevfs_ctrltransfer transfer = { 0x21, 0x09, 0x0200, 0x00, ((kb->fwversion >= 0x300 || IS_V3_OVERRIDE(kb)) ? 2 : 1), 500, ileds };
-    queued_mutex_unlock(dmutex(kb));
-    int res = ioctl(kb->handle - 1, USBDEVFS_CONTROL, &transfer);
-    queued_mutex_lock(dmutex(kb));
-    if(res <= 0) {
-        ckb_err("%s", res ? strerror(errno) : "No data written");
-        if (usb_tryreset(kb) == 0 && countForReset++ < 3) {
-            os_sendindicators(kb);
-        }
-    }
-}
-
 ///
 /// \brief os_inputmain This function is run in a separate thread and will be detached from the main thread, so it needs to clean up its own resources.
 /// \todo This function is a collection of many tasks. It should be divided into several sub-functions for the sake of greater convenience:
@@ -298,23 +91,16 @@ void os_sendindicators(usbdevice* kb) {
 void* os_inputmain(void* context){
     usbdevice* kb = context;
     int fd = kb->handle - 1;
-    ushort vendor = kb->vendor, product = kb->product;
     int index = INDEX_OF(kb, keyboard);
     ckb_info("Starting input thread for %s%d", devpath, index);
 
-    /// Here the actions in detail:
-    ///
-    /// Monitor input transfers on all endpoints for non-RGB devices
-    /// For RGB, monitor all but the last, as it's used for input/output
-    int urbcount = ((IS_LEGACY(vendor, product) || product == P_ST100) ? kb->epcount : (kb->epcount - 1));
-    if (urbcount == 0) {
-        ckb_err("urbcount = 0, so there is nothing to claim in os_inputmain()");
+    if (kb->input_endpoints[0] == 0) {
+        ckb_err("No endpoints claimed in inputmain");
         return 0;
     }
 
     /// Get an usbdevfs_urb data structure and clear it via memset()
-    struct usbdevfs_urb urbs[urbcount];
-    memset(urbs, 0, sizeof(urbs));
+    struct usbdevfs_urb urbs[USB_EP_MAX] = {0};
 
     /// Query udev for wMaxPacketSize on each endpoint, due to certain devices sending more data than the max defined, causing all sorts of issues.
     /// A syspath example would be:
@@ -334,72 +120,86 @@ void* os_inputmain(void* context){
     // Create a list containing them
     struct udev_list_entry* udeventry = udev_enumerate_get_list_entry(enumerate);
 
-    for(int i = 0; i < urbcount; i++){
-        ushort ep = 0x80 | (i + 1);
-
+    int ifcount = 0;
+    do {
         // Move to the next entry in the udev list (skipping the first one).
         struct udev_list_entry* nextentry = udev_list_entry_get_next(udeventry);
         const char* path = udev_list_entry_get_name(nextentry);
         // If there's an underscore, that means we are dealing with udev iterating through endpoints
         // usbX/X-X/X-X:1.0/ep_80
         // ~~~~~~~~~~~~~~~~~~~^
-        if(path[strlen(path) - 3] == '_'){
+        // We only want to iterate through the interfaces
+        size_t pathlen = strlen(path);
+        if(path[pathlen - 3] == '_'){
             ckb_info("Applying udev endpoint workaround for %s", path);
             // Skip the current entry
             udeventry = nextentry;
             nextentry = udev_list_entry_get_next(udeventry);
             path = udev_list_entry_get_name(nextentry);
+            pathlen = strlen(path);
         }
         // Create the path to the endpoint
-        char finalpath[strlen(path)+7];
-        // Copy the base path
-        strcpy(finalpath, path);
-        // Append the endpoint
-        char epstr[7];
-        snprintf(epstr, 7, "/ep_%02x", ep & 0xFF);
-        strcat(finalpath, epstr);
-        // Access it
-        struct udev_device* child = udev_device_new_from_syspath(dev_udev, finalpath);
-        const char* sizehex = udev_device_get_sysattr_value(child, "wMaxPacketSize");
-        // Read its wMaxPacketSize
+        size_t finalpathlen = pathlen + 7;
+        char* finalpath = malloc(finalpathlen);
+        // Try to find any of the wanted endpoints in the current interface
+        // We'll assume that each interface has at most one IN endpoint
         ushort size = 64;
-        if(sizehex)
-            sscanf(sizehex, "%hx", &size);
-        else
-            ckb_warn("Unable to read wMaxPacketSize for %s, assuming 64", epstr);
-#ifdef DEBUG
-        ckb_info("Endpoint path %s has wMaxPacketSize %i", epstr, size);
+        ushort ep = 0;
+        for(int i = 0; (ep = kb->input_endpoints[i]); i++){
+            // Build the path
+            snprintf(finalpath, finalpathlen, "%s/ep_%02hhx", path, ep);
+            // Access it
+            struct udev_device* child = udev_device_new_from_syspath(dev_udev, finalpath);
+            const char* sizehex = udev_device_get_sysattr_value(child, "wMaxPacketSize");
+            // Read its wMaxPacketSize
+            if(sizehex && sscanf(sizehex, "%hx", &size) == 1)
+            {
+                ckb_info("Found EP 0x%hx at %s", ep, finalpath);
+                udev_device_unref(child);
+                break;
+            }
+            udev_device_unref(child);
+        }
+        if(!ep)
+            ckb_fatal("Unable to read wMaxPacketSize for %s, assuming 64", finalpath);
+
+#ifndef NDEBUG
+        ckb_info("Endpoint path %s has wMaxPacketSize %i", finalpath, size);
 #endif
         // Increment the udev list pointer
         udeventry = nextentry;
         // Set the URB parameters
-        urbs[i].buffer_length = size;
-        urbs[i].type = USBDEVFS_URB_TYPE_INTERRUPT;
-        urbs[i].endpoint = ep;
-        urbs[i].buffer = malloc(urbs[i].buffer_length);
-        ioctl(fd, USBDEVFS_SUBMITURB, urbs + i);
+        urbs[ifcount].buffer_length = size;
+        urbs[ifcount].type = USBDEVFS_URB_TYPE_INTERRUPT;
+        urbs[ifcount].endpoint = ep;
+        urbs[ifcount].buffer = malloc(size);
+        ioctl(fd, USBDEVFS_SUBMITURB, urbs + ifcount);
         // Clean up
-        udev_device_unref(child);
-    }
+        free(finalpath);
+        ifcount++;
+    } while (*(kb->input_endpoints + ifcount));
 
     udev_enumerate_unref(enumerate);
     /// The userSpaceFS knows the URBs now, so start monitoring input
     while (1) {
-        struct usbdevfs_urb* urb = 0;
+        struct usbdevfs_urb* urb = NULL;
 
         /// if the ioctl returns something != 0, let's have a deeper look what happened.
         /// Broken devices or shutting down the entire system leads to closing the device and finishing this thread.
-        if (ioctl(fd, USBDEVFS_REAPURB, &urb)) {
-            if (errno == ENODEV || errno == ENOENT || errno == ESHUTDOWN)
+        int res = ioctl(fd, USBDEVFS_REAPURB, &urb);
+        if (res) {
+            int ioctlerrno = errno;
+            wait_until_suspend_processed();
+            if (ioctlerrno == ENODEV || ioctlerrno == ENOENT || ioctlerrno == ESHUTDOWN)
                 // Stop the thread if the handle closes
                 break;
-            else if(errno == EPIPE && urb){
+            else if(ioctlerrno == EPIPE && urb){
                 /// If just an EPIPE ocurred, give the device a CLEAR_HALT and resubmit the URB.
                 ioctl(fd, USBDEVFS_CLEAR_HALT, &urb->endpoint);
                 // Re-submit the URB
                 if(urb)
                     ioctl(fd, USBDEVFS_SUBMITURB, urb);
-                urb = 0;
+                urb = NULL;
             }
             continue;
         }
@@ -415,8 +215,10 @@ void* os_inputmain(void* context){
             process_input_urb(kb, urb->buffer, urb->actual_length, urb->endpoint);
 
             /// Re-submit the URB for the next run.
-            ioctl(fd, USBDEVFS_SUBMITURB, urb);
-            urb = 0;
+            if (ioctl(fd, USBDEVFS_SUBMITURB, urb)) {
+                wait_until_suspend_processed();
+            }
+            urb = NULL;
         }
     }
 
@@ -424,7 +226,7 @@ void* os_inputmain(void* context){
     /// If the endless loop is terminated, clean up by discarding the URBs via ioctl(USBDEVFS_DISCARDURB),
     /// free the URB buffers and return a null pointer as thread exit code.
     ckb_info("Stopping input thread for %s%d", devpath, index);
-    for(int i = 0; i < urbcount; i++){
+    for(int i = 0; i < ifcount; i++){
         ioctl(fd, USBDEVFS_DISCARDURB, urbs + i);
         free(urbs[i].buffer);
     }
@@ -460,18 +262,18 @@ static int usbunclaim(usbdevice* kb, int resetting) {
         ckb_warn("If you still need the device, you will have to restart ckb-next-daemon.");
         return 0;
     }
-    // For RGB keyboards, the kernel driver should only be reconnected to interfaces 0 and 1 (HID), and only if we're not about to do a USB reset.
+    // For NXP devices, the kernel driver should only be reconnected to interfaces 0 and 1 (HID), and only if we're not about to do a USB reset.
     // Reconnecting any of the others causes trouble.
-    if (!resetting) {
-        struct usbdevfs_ioctl ctl = { 0, USBDEVFS_CONNECT, 0 };
+    if (resetting)
+        return 0;
+    struct usbdevfs_ioctl ctl = { 0, USBDEVFS_CONNECT, 0 };
+    for(int i = 0; i < count; i++){
+        // Do not reattach interface 2 if it's an NXP device
+        // FIXME: Is this still an issue?
+        if(i == 2 && kb->protocol == PROTO_NXP)
+            continue;
+        ctl.ifno = i;
         ioctl(handle, USBDEVFS_IOCTL, &ctl);
-        ctl.ifno = 1;
-        ioctl(handle, USBDEVFS_IOCTL, &ctl);
-        // Also reconnect iface #2 (HID) for non-RGB keyboards
-        if(!HAS_FEATURES(kb, FEAT_RGB)){
-            ctl.ifno = 2;
-            ioctl(handle, USBDEVFS_IOCTL, &ctl);
-        }
     }
     return 0;
 }
@@ -509,16 +311,31 @@ void os_closeusb(usbdevice* kb){
 ///
 static int usbclaim(usbdevice* kb){
     int count = kb->epcount;
-#ifdef DEBUG
-    ckb_info("claiming %d endpoints", count);
+#ifndef NDEBUG
+    ckb_info("ckb%d: Claiming %d interfaces", INDEX_OF(kb, keyboard), count);
 #endif // DEBUG
 
+    int retries = 0;
     for(int i = 0; i < count; i++){
-        struct usbdevfs_ioctl ctl = { i, USBDEVFS_DISCONNECT, 0 };
-        ioctl(kb->handle - 1, USBDEVFS_IOCTL, &ctl);
-        if(ioctl(kb->handle - 1, USBDEVFS_CLAIMINTERFACE, &i)) {
-            ckb_err("Failed to claim interface %d: %s", i, strerror(errno));
-            return -1;
+        while (1) {
+            struct usbdevfs_ioctl ctl = { i, USBDEVFS_DISCONNECT, 0 };
+            ioctl(kb->handle - 1, USBDEVFS_IOCTL, &ctl);
+            if(ioctl(kb->handle - 1, USBDEVFS_CLAIMINTERFACE, &i)) {
+                if (errno == EBUSY && retries < 50) {
+                    retries++;
+                    struct timespec sleep_for = { .tv_nsec = 100 * 1000000 };
+                    clock_nanosleep(CLOCK_MONOTONIC, 0, &sleep_for, NULL);
+                    continue;
+                }
+#ifndef NDEBUG
+                if (errno == EBUSY)
+                    ckb_info("Retry timeout after %d seconds", retries / 10);
+#endif // DEBUG
+                ckb_err("Failed to claim interface %d: %s", i, strerror(errno));
+                return -1;
+            } else {
+                break;
+            }
         }
     }
     return 0;
@@ -597,17 +414,17 @@ int os_setupusb(usbdevice* kb) {
     struct udev_device* dev = kb->udev;
     const char* name = udev_device_get_sysattr_value(dev, "product");
     if(name)
-        strncpy(kb->name, name, KB_NAME_LEN);
+        snprintf(kb->name, KB_NAME_LEN, "%s", name);
     strtrim(kb->name);
     const char* serial = udev_device_get_sysattr_value(dev, "serial");
     if(serial)
-        strncpy(kb->serial, serial, SERIAL_LEN);
+        snprintf(kb->serial, SERIAL_LEN, "%s", serial);
     strtrim(kb->serial);
     ///
     /// - Copy firmware version (needed to determine USB protocol)
     const char* firmware = udev_device_get_sysattr_value(dev, "bcdDevice");
     if(firmware)
-        sscanf(firmware, "%hx", &kb->fwversion);
+        sscanf(firmware, "%"SCNx32, &kb->fwversion);
     else
         kb->fwversion = 0;
     int index = INDEX_OF(kb, keyboard);
@@ -621,7 +438,7 @@ int os_setupusb(usbdevice* kb) {
     /// \todo in these modules a pullrequest is outstanding
     ///
     const char* ep_str = udev_device_get_sysattr_value(dev, "bNumInterfaces");
-#ifdef DEBUG
+#ifndef NDEBUG
     ckb_info("Claiming interfaces. name=%s, firmware=%s, ep_str=%s", name, firmware, ep_str);
 #endif //DEBUG
     kb->epcount = 0;
@@ -668,8 +485,8 @@ int usbadd(struct udev_device* dev, ushort vendor, ushort product) {
         ckb_err("Failed to get device path");
         return -1;
     }
-#ifdef DEBUG
-    ckb_info(">>>vendor = 0x%x, product = 0x%x, path = %s, syspath = %s", vendor, product, path, syspath);
+#ifndef NDEBUG
+    ckb_info("Adding device with vendor = 0x%x, product = 0x%x, path = %s, syspath = %s", vendor, product, path, syspath);
 #endif // DEDBUG
 
     for(int i = 0; i < DEV_MAX; i++){
@@ -688,15 +505,7 @@ int usbadd(struct udev_device* dev, ushort vendor, ushort product) {
             // If the mutex is locked then the device is obviously in use, so keep going
             continue;
         }
-        // We can't use IS_CONNECTED() here.
-        // If multiple devices are being opened at once, we will attempt to set up the second one before
-        // we're done with the first one, which means the uinput handles might still be 0.
-        // This will cause IS_CONNECTED to return 0, resulting in overwriting ckb1's kb fields with ckb2's,
-        // leaving ckb1 in an unknown and possibly uninitialised state.
-        //
-        // This happened rarely because dmutex was most likely already locked, so we just skipped over the device
-        // before getting here. (See trylock above.)
-        if(!kb->handle){
+        if(kb->status == DEV_STATUS_DISCONNECTED){
             // Open the sysfs device
             kb->handle = open(path, O_RDWR) + 1;
             if(kb->handle <= 0){
@@ -710,6 +519,7 @@ int usbadd(struct udev_device* dev, ushort vendor, ushort product) {
                 kb->vendor = vendor;
                 kb->product = product;
                 strncpy(kbsyspath[index], syspath, FILENAME_MAX);
+                kb->status = DEV_STATUS_CONNECTING;
                 // Mutex remains locked
                 setupusb(kb);
                 return 0;
@@ -752,8 +562,11 @@ static int usb_add_device(struct udev_device* dev){
         return 1;
 
     for(size_t c = 0; c < N_MODELS; c++){
-        if(models[c].idVendor == vid && models[c].idProduct == pid)
+        if(models[c].idVendor == vid && models[c].idProduct == pid){
+            if(IS_EXPERIMENTAL(vid, pid) && !enable_experimental)
+                continue;
             return usbadd(dev, models[c].idVendor, models[c].idProduct);
+        }
     }
 
     return 1;
@@ -820,6 +633,90 @@ static void udev_enum(){
 
 _Atomic int suspend_run = 1;
 
+///
+/// \brief Gracefully resume from suspend.
+static void graceful_suspend_resume() {
+    // Wait for the kernel to finish reattaching drivers. Avoids race
+    // conditions (kernel bug) and situations where the usbhid driver is
+    // attached after we already believe to be in control of the device.
+    // Retry for at most 5 seconds.
+    // XXX: Is 5s too much? It might be possible for interfaces to be left
+    // unclaimed indefinitely.
+    for (int try = 0; suspend_run && try < 50; try++) {
+        bool drivers_attached = true;
+        for (int i = 1; i < DEV_MAX; i++) {
+            usbdevice* kb = keyboard + i;
+            queued_mutex_lock(dmutex(kb));
+            if (kb->status == DEV_STATUS_CONNECTED && kb->active) {
+                for (int j = 0; j < kb->epcount; j++) {
+                    // Replicates the logic from usbunclaim() - ignore
+                    // interfaces other than 0 and 1.
+                    if (j == 2 && kb->protocol == PROTO_NXP)
+                        continue;
+
+                    struct usbdevfs_getdriver query = {
+                        .interface = j,
+                    };
+                    if (ioctl(kb->handle - 1, USBDEVFS_GETDRIVER, &query)) {
+                        if (errno == ENODATA) {
+                            // No driver attached yet, so keep waiting.
+                            drivers_attached = false;
+                        }
+                        // Ignore other errors.
+                    }
+                }
+            }
+            queued_mutex_unlock(dmutex(kb));
+        }
+        if (drivers_attached) {
+#ifndef NDEBUG
+            ckb_info("graceful_suspend_resume: drivers_attached");
+#endif // DEBUG
+            break;
+        } else {
+            struct timespec sleep_for = { .tv_nsec = 100 * 1000000 };
+            clock_nanosleep(CLOCK_MONOTONIC, 0, &sleep_for, NULL);
+        }
+    }
+    // If some interfaces are now claimed by a driver other than usbfs (or
+    // none), attemt to reclaim them.
+    for (int i = 1; i < DEV_MAX; i++) {
+        usbdevice* kb = keyboard + i;
+        queued_mutex_lock(dmutex(kb));
+        if (kb->status == DEV_STATUS_CONNECTED && kb->active) {
+            bool needs_reclaim = false;
+            for (int j = 0; j < kb->epcount; j++) {
+                struct usbdevfs_getdriver query = {
+                    .interface = j,
+                };
+                if (!ioctl(kb->handle - 1, USBDEVFS_GETDRIVER, &query)) {
+                    if (strcmp("usbfs", query.driver)) {
+#ifndef NDEBUG
+                        ckb_info("ckb%d: USBDEVFS_GETDRIVER on interface %d gave %s, reclaiming", i, j, query.driver);
+#endif // DEBUG
+                        needs_reclaim = true;
+                        break;
+                    }
+                } else {
+                    if (errno == ENODATA) {
+#ifndef NDEBUG
+                        ckb_info("ckb%d: USBDEVFS_GETDRIVER on interface %d gave ENODATA, reclaiming", i, j);
+#endif // DEBUG
+                        needs_reclaim = true;
+                        break;
+                    }
+                }
+            }
+            if (needs_reclaim) {
+                usbclaim(kb);
+                // Ignore errors for now. reactivate_devices() should
+                // try to reset the device.
+            }
+        }
+        queued_mutex_unlock(dmutex(kb));
+    }
+}
+
 // While POSIX says we may need a runtime test, on Linux the clock
 // was added immediately after the define was introduced, so a
 // compile time check should be enough
@@ -840,22 +737,59 @@ static time_t get_clock_monotonic_seconds() {
     return timespec_var.tv_sec;
 }
 
+static pthread_mutex_t suspend_check_mutex = PTHREAD_MUTEX_INITIALIZER;
+static pthread_cond_t suspend_check_cond = PTHREAD_COND_INITIALIZER;
+static time_t prev_suspend_check_time;
+
 void* suspend_check() {
-    time_t prev_time = get_clock_monotonic_seconds();
-    if(!prev_time)
+    time_t current_time = get_clock_monotonic_seconds();
+    if(!current_time)
         return NULL;
+    pthread_mutex_lock(&suspend_check_mutex);
+    prev_suspend_check_time = current_time;
+    pthread_mutex_unlock(&suspend_check_mutex);
+    bool woke_up = false;
     while(suspend_run){
         clock_nanosleep(CLOCK_MONOTONIC, 0, &(struct timespec) {.tv_sec = 2}, NULL);
-        time_t current_time = get_clock_monotonic_seconds(NULL);
+        current_time = get_clock_monotonic_seconds();
 
-        if(prev_time + 4 < current_time)
+        if (prev_suspend_check_time + 4 < current_time) {
+            graceful_suspend_resume();
+            // Some time might have passed, get a fresh timestamp so other
+            // threads can continue
+            current_time = get_clock_monotonic_seconds();
+            woke_up = true;
+        }
+
+        pthread_mutex_lock(&suspend_check_mutex);
+        prev_suspend_check_time = current_time;
+        pthread_cond_broadcast(&suspend_check_cond);
+        pthread_mutex_unlock(&suspend_check_mutex);
+
+        if(woke_up){
             reactivate_devices();
-
-        prev_time = current_time;
+            woke_up = false;
+        }
     }
     return NULL;
 }
 
+void wait_until_suspend_processed() {
+    // Prevent other threads from doing anything after waking from suspend and
+    // before the suspend thread has finished reactivate_devices().
+    // If necessary, this could be optimized by first checking an _Atomic
+    // representing prev_suspend_check_time instead of locking the shared mutex
+    // directly.
+
+    time_t current_time = get_clock_monotonic_seconds();
+    pthread_mutex_lock(&suspend_check_mutex);
+    while (suspend_run && prev_suspend_check_time + 4 < current_time) {
+        pthread_cond_wait(&suspend_check_cond, &suspend_check_mutex);
+        current_time = get_clock_monotonic_seconds();
+    }
+    pthread_mutex_unlock(&suspend_check_mutex);
+}
+
 /// \brief .
 ///
 /// \brief usbmain is called by main() after setting up all other stuff.
@@ -869,12 +803,12 @@ int usbmain(){
         ckb_fatal("Failed to initialize udev in usbmain(), usb_linux.c");
         return -1;
     }
-    
+
     // Create thread that detects system suspend
     pthread_t suspend_thread;
     pthread_create(&suspend_thread, NULL, suspend_check, NULL);
     pthread_setname_np(suspend_thread, "suspend");
-    
+
     ///
     /// Enumerate all currently connected devices
     udev_enum();
@@ -888,7 +822,7 @@ int usbmain(){
     // Get an fd for the monitor
     int fd = udev_monitor_get_fd(monitor);
     fd_set fds;
-    
+
     while(udev){
         FD_ZERO(&fds);
         FD_SET(fd, &fds);
@@ -937,7 +871,7 @@ int usbmain(){
     udev_monitor_unref(monitor);
     suspend_run = 0;
     pthread_join(suspend_thread, NULL);
-    
+
     return 0;
 }
 
