@@ -1,6 +1,5 @@
 #include <cmath>
 #include <QApplication>
-#include <QDateTime>
 #include <QDebug>
 #include <QDir>
 #include <QUrl>
@@ -194,7 +193,7 @@ bool AnimScript::load(){
             QMap<QString, QVariant> preset;
             for(int i = 2; i < count; i++){
                 // Scan name/value setting pairs
-                QString setting = components.at(i);
+                const QString& setting = components.at(i);
                 QStringList sComponents = setting.split("=");
                 if(sComponents.count() != 2)
                     continue;
