@@ -2,7 +2,6 @@
 #include <QSet>
 #include <QUrl>
 #include <QMutex>
-#include <QDateTime>
 #include <QDebug>
 #include "kb.h"
 #include "kbmanager.h"
@@ -17,10 +16,9 @@ int Kb::_frameRate = 30, Kb::_scrollSpeed = 0;
 bool Kb::_dither = false, Kb::_mouseAccel = true;
 
 Kb::Kb(QObject *parent, const QString& path) :
-    QThread(parent), features("N/A"), firmware("N/A"), pollrate("N/A"), monochrome(false), hwload(false), adjrate(false),
-    devpath(path), cmdpath(path + "/cmd"), notifyPath(path + "/notify1"), macroPath(path + "/notify2"),
-    _currentProfile(0), _currentMode(0), _model(KeyMap::NO_MODEL),
-    lastAutoSave(QDateTime::currentMSecsSinceEpoch()),
+    QThread(parent), features(QStringList()), pollrate("N/A"), monochrome(false), hwload(false), adjrate(false), firmware(),
+    batteryTimer(0), batteryIcon(0), showBatteryIndicator(false), devpath(path), cmdpath(path + "/cmd"), notifyPath(path + "/notify1"), macroPath(path + "/notify2"),
+    _currentProfile(0), _currentMode(0), _model(KeyMap::NO_MODEL), batteryLevel(0), batteryStatus(BatteryStatus::BATT_STATUS_UNKNOWN),
     _hwProfile(0), prevProfile(0), prevMode(0),
     cmd(cmdpath), notifyNumber(1), macroNumber(2), _needsSave(false), _layout(KeyMap::NO_LAYOUT), _maxDpi(0),
     deviceIdleTimer()
@@ -31,16 +29,15 @@ Kb::Kb(QObject *parent, const QString& path) :
     // Get the features, model, serial number, FW version (if available), poll rate (if available), and layout from /dev nodes
     QFile ftpath(path + "/features"), mpath(path + "/model"), spath(path + "/serial"), fwpath(path + "/fwversion"), ppath(path + "/pollrate"), prodpath(path + "/productid"), hwlayoutPath(path + "/layout"), dpiPath(path + "/dpi");
     if (ftpath.open(QIODevice::ReadOnly)){
-        features = ftpath.read(1000);
-        features = features.trimmed();
+        QString featurestr = ftpath.read(1000).trimmed();
         ftpath.close();
         // Read model from features (first word: vendor, second word: product)
-        QStringList list = features.split(" ");
-        if(list.length() < 2)
+        features = featurestr.split(" ");
+        if(features.length() < 2)
             return;
-        _model = KeyMap::getModel(list[1]);
+        _model = KeyMap::getModel(features[1]);
         if(_model == KeyMap::NO_MODEL) {
-            qDebug() << "could not find valid model information:" << list[1] << "produced" << _model;
+            qDebug() << "could not find valid model information:" << features[1] << "produced" << _model;
             return;
         }
     } else {
@@ -66,12 +63,11 @@ Kb::Kb(QObject *parent, const QString& path) :
     }
     if (usbSerial == "")
         usbSerial = "Unknown-" + usbModel;
-    if (features.contains("fwversion") && fwpath.open(QIODevice::ReadOnly)) {
-        firmware = fwpath.read(100);
-        firmware = QString::number(firmware.trimmed().toInt() / 100., 'f', 2);
+    if (fwpath.open(QIODevice::ReadOnly)) {
+        firmware.parse(fwpath.read(100));
         fwpath.close();
         if (prodpath.open(QIODevice::ReadOnly)) {
-            productID = prodpath.read(4).toUShort(0, 16);
+            productID = prodpath.read(4).toUShort(nullptr, 16);
             // qInfo() << "ProductID of device is" << productID;
         } else {
             qCritical() << "could not open" << prodpath.fileName();
@@ -123,6 +119,18 @@ Kb::Kb(QObject *parent, const QString& path) :
     cmd.write(QString("notifyon %1\n").arg(notifyNumber).toLatin1());
     cmd.flush();
 
+    if(features.contains("battery")){
+        batteryIcon = new BatteryStatusTrayIcon(usbModel, this);
+        updateBattery();
+        batteryTimer = new QTimer(this);
+        connect(batteryTimer, &QTimer::timeout, this, &Kb::updateBattery);
+        connect(this, &Kb::batteryChanged, batteryIcon, &BatteryStatusTrayIcon::setBattery);
+        if(showBatteryIndicator)
+            batteryIcon->show();
+        batteryTimer->setInterval(10000);
+        batteryTimer->start();
+    }
+
     // Again, find an available notification node for macro definition
     // (if none is found, take notify2)
     {
@@ -169,6 +177,9 @@ Kb::~Kb(){
     // Save settings first
     save();
 
+    delete batteryTimer;
+    delete batteryIcon;
+
     // remove the notify channel from the list of notifyPaths.
     ///< \todo I don't think, that notifypaths is used somewhere. So why do we have it?
     /// If we do not need it, searching for an ununsed notify channel can easy be refactored to a private member function.
@@ -176,16 +187,16 @@ Kb::~Kb(){
 
     // Kill notification thread and remove node
     activeDevices.remove(this);
-    if(!isOpen()){
-        terminate();
-        wait(1000);
-        return;
-    }
-    if(notifyNumber > 0)
+    if(cmd.isOpen() && notifyNumber > 0){
         cmd.write(QString("idle\nnotifyoff %1\n").arg(notifyNumber).toLatin1());
-    cmd.flush();
-    terminate();
-    wait(1000);
+        // Manually flush so that the daemon closes the notify pipe and the thread can gracefully stop
+        cmd.flush();
+    }
+    if(!wait(1000)){
+        terminate();
+        qDebug() << "Second wait returned" << wait(1000);
+    }
+    // This does nothing if cmd isn't open.
     cmd.close();
 }
 
@@ -229,6 +240,11 @@ void Kb::updateLayout(bool stop){
     emit infoUpdated();
 }
 
+void Kb::updateBattery(){
+    cmd.write(QString("@%1 get :battery\n").arg(notifyNumber).toLatin1());
+    cmd.flush();
+}
+
 void Kb::dither(bool newDither){
     if(newDither == _dither)
         return;
@@ -283,6 +299,7 @@ void Kb::load(){
                 newCurrentProfile = profile;
         }
     }
+    showBatteryIndicator = settings.value("batteryIndicator", true).toBool();
     if(newCurrentProfile)
         setCurrentProfile(newCurrentProfile);
     else {
@@ -300,7 +317,7 @@ void Kb::load(){
         else if(map.model() == KeyMap::NIGHTSWORD)
             demoProfile = ":/txt/demoprofile_nightsword.ini";
         QSettings demoSettings(demoProfile, QSettings::IniFormat, this);
-        CkbSettings cSettings(demoSettings);
+        CkbDemoSettings cSettings(demoSettings);
         KbProfile* demo = new KbProfile(this, map, cSettings, "{BA7FC152-2D51-4C26-A7A6-A036CC93D924}");
         _profiles.append(demo);
         setCurrentProfile(demo);
@@ -325,14 +342,12 @@ void Kb::save(){
     settings.setValue("CurrentProfile", currentGuid);
     settings.setValue("Profiles", guids.trimmed());
     settings.setValue("hwLayout", KeyMap::getLayout(_layout));
+    settings.setValue("batteryIndicator", showBatteryIndicator);
 }
 
 void Kb::autoSave(){
-    quint64 now = QDateTime::currentMSecsSinceEpoch();
-    if(needsSave() && now >= lastAutoSave + 15 * 1000 && !CkbSettings::isBusy()){
+    if(needsSave() && !CkbSettings::isBusy())
         save();
-        lastAutoSave = now;
-    }
 }
 
 void Kb::hwSave(){
@@ -505,9 +520,10 @@ void Kb::run(){
     }
     QMutexLocker locker(&notifyPathMutex);
     notifyPaths.remove(notifyPath);
+    qDebug() << "Notify thread returning. Read" << line.length() << "isOpen()" << notify.isOpen();
 }
 
-void Kb::readNotify(QString line){
+void Kb::readNotify(const QString& line){
     QStringList components = line.trimmed().split(" ");
     if(components.count() < 2)
         return;
@@ -524,6 +540,18 @@ void Kb::readNotify(QString line){
             mode->bind()->keyEvent(keyName, keyPressed);
         }
         deviceIdleTimer.start();
+    } else if (components[0] == "battery"){
+        QStringList bComponents = components[1].split(':');
+        if(bComponents.length() != 2)
+            return;
+        // Convert battery values into human readable text
+        bool ok, ok2;
+        uint newBatteryLevel = bComponents[0].toUInt(&ok), newBatteryStatus = bComponents[1].toUInt(&ok2);
+        if(!ok || !ok2 || newBatteryStatus >= BatteryStatus::BATT_STATUS_INVALID || (batteryLevel == newBatteryLevel && batteryStatus == newBatteryStatus))
+            return;
+        batteryLevel = newBatteryLevel;
+        batteryStatus = static_cast<BatteryStatus>(newBatteryStatus);
+        emit batteryChanged(batteryLevel, batteryStatus);
     } else if(components[0] == "i"){
         // Indicator event
         QString i = components[1];
@@ -710,7 +738,7 @@ void Kb::readNotify(QString line){
                 if(dpi.length() != 2)
                     continue;
                 QStringList xy = dpi[1].split(',');
-                int x, y;
+                int x = 0, y = 0;
                 bool off = false;
                 if(xy.length() < 2){
                     // If the right side only has one parameter, set both X and Y
@@ -800,7 +828,11 @@ void Kb::setCurrentProfile(KbProfile* profile){
     emit profileAboutToChange();
     _currentProfile = profile;
     emit profileChanged();
-    setCurrentMode(profile->currentMode());
+    // Hack to prevent crash when switching to HW mode on first start with no config file.
+    // It happens when called by KbWidget::on_profileBox_activated().
+    // The KbWidget event will re-call this after the currentMode has properly been set.
+    if(profile->currentMode())
+        setCurrentMode(profile->currentMode());
 }
 
 void Kb::setCurrentMode(KbMode* mode){
@@ -822,7 +854,6 @@ KeyMap::Layout Kb::getCurrentLayout(){
     return _layout;
 }
 
-void Kb::setPollRate(QString poll)
-{
+void Kb::setPollRate(const QString& poll){
     cmd.write(QString("\npollrate %1\n").arg(poll).toLatin1());
 }
