@@ -4,9 +4,32 @@
 #include <QObject>
 #include <QFile>
 #include <QThread>
+#include <QTimer>
 #include "kbprofile.h"
 #include <QElapsedTimer>
 #include <limits>
+#include "batterysystemtrayicon.h"
+#include "ckbversionnumber.h"
+
+struct firmware_t {
+    CkbVersionNumber app;
+    CkbVersionNumber bld;
+    CkbVersionNumber radioapp;
+    CkbVersionNumber radiobld;
+    void parse(const QString& str){
+        // KeepEmptyParts is the default
+        QVector<QStringRef> split = str.leftRef(str.size()-1).split(QChar('\n'));
+        // Old < 0.5.0 format (NXP and legacy)
+        if(split.size() == 1){
+            app = CkbVersionNumber(split.at(0).toString());
+        } else if (split.size() >= 4) {
+            app = CkbVersionNumber(split.at(0).toString());
+            bld = CkbVersionNumber(split.at(1).toString());
+            radioapp = CkbVersionNumber(split.at(2).toString());
+            radiobld = CkbVersionNumber(split.at(3).toString());
+        }
+    }
+};
 
 // Class for managing devices
 class Kb : public QThread
@@ -16,11 +39,13 @@ public:
     // USB model and serial number
     QString usbModel, usbSerial;
     // Device information
-    QString features, firmware, pollrate;
+    QStringList features;
+    QString pollrate;
     bool monochrome;
     ushort productID;
     bool hwload;
     bool adjrate;
+    firmware_t firmware;
 
     // Keyboard model
     inline KeyMap::Model    model() const                       { return _model; }
@@ -92,6 +117,14 @@ public:
 
     KeyMap::Layout getCurrentLayout();
 
+    // Battery polling timer
+    QTimer* batteryTimer;
+
+    // Battery status icon
+    BatteryStatusTrayIcon* batteryIcon;
+
+    bool showBatteryIndicator;
+
     //////////
     /// For usage with macro definions, these two params must only be readable.
     /// So there are no setters.
@@ -112,7 +145,7 @@ public:
     inline QString getMacroPath () { return macroPath; }
 
     inline ushort getMaxDpi () {return _maxDpi; }
-    void setPollRate(QString poll);
+    void setPollRate(const QString& poll);
 
     // The valid check is done because we don't start the timer on purpose in the constructor.
     // This is done so that when a new device is plugged in while the lights are off, it doesn't suddenly return a really low value and wake everything up.
@@ -127,6 +160,7 @@ signals:
     // Profile/mode updates
     void profileAdded();
     void profileRenamed();
+    void batteryChanged(uint batteryLevel, BatteryStatus batteryStatus);
 
     void profileChanged();
     void profileAboutToChange();
@@ -145,10 +179,11 @@ public slots:
 
 private slots:
     // Processes lines read from the notification node
-    void readNotify(QString line);
+    void readNotify(const QString& line);
 
     void deleteHw();
     void deletePrevious();
+    void updateBattery();
 
 private:
     // Following methods should only be used by KbManager
@@ -183,12 +218,14 @@ private:
 
     KeyMap::Model   _model;
 
+    uint batteryLevel;
+    BatteryStatus batteryStatus;
+
     // Indicator light state
     bool iState[KbPerf::HW_I_COUNT];
 
     // CkbSettings path
     QString prefsPath;
-    quint64 lastAutoSave;
 
     // Current firmware update file
     QString fwUpdPath;
