@@ -1,5 +1,5 @@
 #include <cmath>
-#include <QDateTime>
+#include "monotonicclock.h"
 #include <QSet>
 #include "kblight.h"
 #include "kbmode.h"
@@ -108,7 +108,7 @@ void KbLight::dimming(int newDimming, bool noSave){
 KbAnim* KbLight::addAnim(const AnimScript *base, const QStringList &keys, const QString& name, const QMap<QString, QVariant>& preset){
     // Stop and restart all existing animations
     stopPreview();
-    quint64 timestamp = QDateTime::currentMSecsSinceEpoch();
+    quint64 timestamp = MonotonicClock::msecs();
     foreach(KbAnim* anim, _animList){
         anim->stop();
         anim->trigger(timestamp);
@@ -132,7 +132,7 @@ KbAnim* KbLight::addAnim(const AnimScript *base, const QStringList &keys, const
 void KbLight::previewAnim(const AnimScript* base, const QStringList& keys, const QMap<QString, QVariant>& preset){
     if(_previewAnim)
         stopPreview();
-    quint64 timestamp = QDateTime::currentMSecsSinceEpoch();
+    quint64 timestamp = MonotonicClock::msecs();
     // Load the new animation and set preset parameters
     KbAnim* anim = new KbAnim(this, _map, "", keys, base);
     QMapIterator<QString, QVariant> i(preset);
@@ -155,7 +155,7 @@ void KbLight::stopPreview(){
 
 KbAnim* KbLight::duplicateAnim(KbAnim* oldAnim){
     // Stop and restart all existing animations
-    quint64 timestamp = QDateTime::currentMSecsSinceEpoch();
+    quint64 timestamp = MonotonicClock::msecs();
     foreach(KbAnim* anim, _animList){
         anim->stop();
         anim->trigger(timestamp);
@@ -185,7 +185,7 @@ bool KbLight::isStarted(){
 }
 
 void KbLight::restartAnimation(){
-    quint64 timestamp = QDateTime::currentMSecsSinceEpoch();
+    quint64 timestamp = MonotonicClock::msecs();
     foreach(KbAnim* anim, _animList){
         anim->stop();
         anim->trigger(timestamp);
@@ -197,11 +197,11 @@ void KbLight::restartAnimation(){
 void KbLight::animKeypress(const QString& key, bool down){
     foreach(KbAnim* anim, _animList){
         if(anim->keys().contains(key))
-            anim->keypress(key, down, QDateTime::currentMSecsSinceEpoch());
+            anim->keypress(key, down, MonotonicClock::msecs());
     }
     if(_previewAnim){
         if(_previewAnim->keys().contains(key))
-            _previewAnim->keypress(key, down, QDateTime::currentMSecsSinceEpoch());
+            _previewAnim->keypress(key, down, MonotonicClock::msecs());
     }
 }
 
@@ -212,7 +212,7 @@ void KbLight::open(){
     activeLights.insert(this);
     if(_start)
         return;
-    quint64 timestamp = QDateTime::currentMSecsSinceEpoch();
+    quint64 timestamp = MonotonicClock::msecs();
     foreach(KbAnim* anim, _animList)
         anim->trigger(timestamp);
     if(_previewAnim)
@@ -279,13 +279,13 @@ static float sToL(float srgb){
     srgb /= 255.f;
     if(srgb <= 0.04045f)
         return srgb / 12.92f;
-    return pow((srgb + 0.055f) / 1.055f, 2.4f);
+    return std::pow((srgb + 0.055f) / 1.055f, 2.4f);
 }
 
 static float lToS(float linear){
     if(linear <= 0.0031308f)
         return 12.92f * linear * 255.f;
-    return (1.055f * pow(linear, 1.f / 2.4f) - 0.055f) * 255.f;
+    return (1.055f * std::pow(linear, 1.f / 2.4f) - 0.055f) * 255.f;
 }
 
 // Convert RGB to monochrome
@@ -297,7 +297,7 @@ QRgb monoRgb(float r, float g, float b){
     r = sToL(r);
     g = sToL(g);
     b = sToL(b);
-    int value = round(lToS((r + g + b) / 3.f));
+    int value = std::round(lToS((r + g + b) / 3.f));
     return qRgb(value, value, value);
 }
 
@@ -309,7 +309,7 @@ void KbLight::frameUpdate(QFile& cmd, bool monochrome){
     rebuildBaseMap();
     _animMap = _colorMap;
     // Advance animations
-    quint64 timestamp = QDateTime::currentMSecsSinceEpoch();
+    quint64 timestamp = MonotonicClock::msecs();
     foreach(KbAnim* anim, _animList)
         anim->blend(_animMap, timestamp);
     if(_previewAnim)
@@ -319,11 +319,13 @@ void KbLight::frameUpdate(QFile& cmd, bool monochrome){
     if(_animMap == _lastFrameAnimMap && _indicatorMap == _lastFrameIndicatorMap && _lastFrameDimming == _dimming && !_forceFrame)
         return;
 
-     _forceFrame = false;
+    // This is used to prevent spamming "rgb 000000" to the daemon when the lights are off
+    const bool lastFrameOrForce = _lastFrameDimming != _dimming || _forceFrame;
 
     _lastFrameAnimMap = _animMap;
     _lastFrameIndicatorMap = _indicatorMap;
     _lastFrameDimming = _dimming;
+    _forceFrame = false;
 
     int count = _animMap.count();
     QRgb* colors = _animMap.colors();
@@ -342,9 +344,9 @@ void KbLight::frameUpdate(QFile& cmd, bool monochrome){
                 float g2 = qGreen(rgb2);
                 float b2 = qBlue(rgb2);
                 float a2 = qAlpha(rgb2) / 255.f;
-                r = round(r2 * a2 + r * (1.f - a2));
-                g = round(g2 * a2 + g * (1.f - a2));
-                b = round(b2 * a2 + b * (1.f - a2));
+                r = std::round(r2 * a2 + r * (1.f - a2));
+                g = std::round(g2 * a2 + g * (1.f - a2));
+                b = std::round(b2 * a2 + b * (1.f - a2));
             }
             // If monochrome mode is active, average the channels to get a grayscale image
             if(monochrome)
@@ -361,7 +363,7 @@ void KbLight::frameUpdate(QFile& cmd, bool monochrome){
     }
 
     // If brightness is at 0%, turn off lighting entirely
-    if(_dimming == 3){
+    if(_dimming == 3 && lastFrameOrForce){
         cmd.write("rgb 000000\n");
         return;
     }
@@ -378,9 +380,9 @@ void KbLight::frameUpdate(QFile& cmd, bool monochrome){
             r *= light;
             g *= light;
             b *= light;
-            r = round(lToS(r));
-            g = round(lToS(g));
-            b = round(lToS(b));
+            r = std::round(lToS(r));
+            g = std::round(lToS(g));
+            b = std::round(lToS(b));
             rgb = qRgb(r, g, b);
         }
     }
