@@ -75,6 +75,15 @@ public:
         M95,
         IRONCLAW,
         NIGHTSWORD,
+        DARKCORE,
+        IRONCLAW_W,
+        KATARPROXT,
+        K95L,
+        GLAIVEPRO,
+        M55,
+        K60,
+        K57_WL,
+        K63_WL,
         _MODEL_MAX
     };
     // Key layouts (ordered alphabetically by name)
@@ -100,7 +109,7 @@ public:
         _LAYOUT_MAX
     };
     // Human-readable names of each layout
-    static QList<QPair<int, QString> > layoutNames(QString layout);
+    static QList<QPair<int, QString> > layoutNames(const QString& layout);
     // ISO (105-key) or ANSI (104-key)?
     inline static bool  isISO(Layout layout)    { return layout != US && layout != US_DVORAK && layout != PL; }
     inline bool         isISO() const           { return isISO(keyLayout); }
@@ -119,7 +128,7 @@ public:
     // Type of device
     inline static bool  isKeyboard(Model model)     { return !isMouse(model) && !isMousepad(model) && !isHeadsetStand(model) && model != NO_MODEL; }
     inline bool         isKeyboard() const          { return isKeyboard(keyModel); }
-    inline static bool  isMouse(Model model)        { return model == M65 || model == SABRE || model == SCIMITAR || model == HARPOON || model == GLAIVE || model == KATAR || model == M65E || model == M95 || model == IRONCLAW || model == NIGHTSWORD; }
+    inline static bool  isMouse(Model model)        { return model == M55 || model == M65 || model == SABRE || model == SCIMITAR || model == HARPOON || model == GLAIVE || model == KATAR || model == KATARPROXT || model == M65E || model == M95 || model == IRONCLAW || model == NIGHTSWORD || model == DARKCORE || model == IRONCLAW_W || model == GLAIVEPRO; }
     inline bool         isMouse() const             { return isMouse(keyModel); }
     inline static bool  isMousepad(Model model)     { return model == POLARIS; }
     inline bool         isMousepad() const          { return isMousepad(keyModel); }
